#include "constants.fxh"
#include "lighting.fxh"

struct vsIn
{
  float4  Position    	: POSITION;
#ifdef VERTEX_COLOR
  FIXED4 VtxColor    	  : COLOR0;
#endif
  half2  Tex0        	: TEXCOORD0;
#if defined(RT_LIGHTMAP) && !defined(RT_CUSTOM_LIGHTMAP)
  half2  Tex1        	: TEXCOORD1;
#endif

  half3  Normal      	: NORMAL;
#ifdef TANGENT_BASIS
  half3  Tangent     	: TANGENT;
  half3  Binormal    	: BINORMAL;
#endif
};

struct vsOut
{
  float4 HPosition			: POSITION;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  FIXED4 VtxColor			: COLOR0;
#endif

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW
  FIXED4 FogColor    		: COLOR1;
 #else	
  FIXED FogColor    		  : FOG;
 #endif	
#endif

  half4 Tex0				: TEXCOORD0;
#if RT_PROFILE <= QUALITY_LOW
  FIXED4 Diffuse			: COLOR0;
#endif

  half3 WorldNormal			: TEXCOORD3;
#if defined(TANGENT_BASIS) && RT_PROFILE > QUALITY_VERYLOW
  half3 WorldTangent			: TEXCOORD4;
  half3 WorldBinormal		 	: TEXCOORD5;
#endif  
};


half3 GetLDiffuse(const float3 vWorldPos, const float3 vWorldNormal)
{
  half3 vDiffuseAcc = half3(0,0,0);

  float4 WorldLightPos = LGetPosition(0);
  half3 vVertexSpaceLightPos = WorldLightPos.xyz - vWorldPos;
  half3 vDirToLight = normalize(vVertexSpaceLightPos);
  float fNdotL = dot(vDirToLight.xyz, vWorldNormal);
  vDiffuseAcc.rgb = saturate(fNdotL) * LGetDiffuse(0);

  return vDiffuseAcc;
}

void LLighting(inout vsOut Out, const in vsIn In, const float3 vWorldPos, const float3 vWorldNormal)
{
#if RT_PROFILE <= QUALITY_LOW    
    Out.Diffuse = float4(0,0,0,1);
  #ifdef RT_NUM_LIGHTS
    Out.Diffuse.rgb += GetLDiffuse(vWorldPos, vWorldNormal.xyz);
  #endif
  #if defined(VERTEX_COLOR)
   #if !defined(RT_STATIC_LIGHTING)
    Out.Diffuse.rgb *= In.VtxColor.rgb;
   #endif	
    Out.Diffuse.a = In.VtxColor.a;
  #endif
  
  #ifndef RT_LIGHTMAP
    Out.Diffuse.rgb += F_WorldAmbient.rgb;
  #endif
#endif
}

vsOut BuildVS(in vsIn In)
{
  vsOut Out = (vsOut)0;
  
  Out.WorldNormal.xyz = 0;

  float4 Po = float4(In.Position.xyz, 1);
  float3 vWorldPos = MUL_VEC_MAT(Po, PI_mW).xyz;
  float3 vWorldNormal = normalize(MUL_VEC_MAT(In.Normal.xyz, (float3x3)PI_mW).xyz);

  Out.Tex0.xy = In.Tex0.xy;
#if defined(RT_LIGHTMAP) || defined(SECONDARY_DIFFUSE)
  #if defined(RT_CUSTOM_LIGHTMAP)
    Out.Tex0.zw = (vWorldPos.xz - LightmapCoords.xy) * LightmapCoords.z;
    Out.Tex0.zw += LightmapCoords.w;
    Out.Tex0.w = 1 - Out.Tex0.w;
  #elif defined(RT_LIGHTMAP)
    Out.Tex0.zw = In.Tex1.xy;
  #endif	
#endif

#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  Out.VtxColor = In.VtxColor;
#endif

  float4 vWPos = float4(vWorldPos, 1);
  Out.HPosition = MUL_VEC_MAT(vWPos, C_VP);

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW && !defined(OPENGL)
  float3 WorldView = MAT_POS(C_VI) - vWorldPos;
  Out.FogColor = Fog(length(WorldView.xyz));
 #else
  Out.FogColor = Fog(Out.HPosition.w);
 #endif	
#endif

#if RT_PROFILE <= QUALITY_LOW
  LLighting(Out, In, vWorldPos, vWorldNormal);
#endif

  return Out;
}

FIXED4 BuildPS(in vsOut IN) : COLOR
{
  half3 vNormal = IN.WorldNormal.xyz;
  half2 baseTC = IN.Tex0.xy;
  half2 lmTC = IN.Tex0.zw;

  half3 mTangentToWS0, mTangentToWS1, mTangentToWS2;
#if defined (TANGENT_BASIS) && RT_PROFILE > QUALITY_VERYLOW
  mTangentToWS0 = IN.WorldTangent.xyz;  
  mTangentToWS1 = IN.WorldBinormal.xyz;  
  mTangentToWS2 = IN.WorldNormal.xyz;  
#endif

  FIXED4 cDiffuseMap = tex2D(ColorSampler, baseTC);

  FIXED fAlpha = cDiffuseMap.a;
  FIXED4 cVertexColor = 1;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  cVertexColor = IN.VtxColor;
  fAlpha *= cVertexColor.a;
#endif

#if defined(SECONDARY_DIFFUSE) && defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  half4 vSecondColor = tex2D(SecondColorSampler, lmTC);
  float fA = cVertexColor.a;
  cDiffuseMap = lerp(cDiffuseMap, vSecondColor, fA);
#endif

#if RT_PROFILE <= QUALITY_LOW
  FIXED3 cDiffuseAcc = IN.Diffuse.rgb;
#else
  FIXED3 cDiffuseAcc = 0;
#endif
  FIXED3 cAmbientAcc = 0;
  FIXED3 cSpecularAcc = 0;

#if RT_PROFILE > QUALITY_LOW
  cAmbientAcc = F_WorldAmbient.rgb;
#endif
#if defined(RT_LIGHTMAP)
  cAmbientAcc = lerp(tex2D(LightMapSampler, lmTC.xy).rgb * 2.h, float3(1,1,1), fAlpha);
#endif

  FIXED4 cOut;
  cOut.xyz = (cAmbientAcc.xyz + cDiffuseAcc.xyz) * cDiffuseMap.xyz * cVertexColor.rgb + cSpecularAcc;
  cOut.w = fAlpha;

#if defined(RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  frag_fog_setup(IN.FogColor, cOut);
#endif

#if defined (RT_ENVMAP) && RT_PROFILE > QUALITY_VERYLOW
  //cOut.xyz = cDiffuseMap.a;
#endif
#if defined(RT_LIGHTMAP)
  //cOut.xyz = fAlpha;
#endif

  return cOut;
}

technique Default
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE BuildVS();
    PixelShader = compile PS_SHADER_TYPE BuildPS();
    DEFAULT_STATES
  }
}
