#include "constants.fxh"
#include "lighting.fxh"

uniform float4 CarColor = {0.0f, 0.0f, 0.0f, 1.0f};
uniform float ScrollSpeed = 0.2f;

struct vsIn
{
  float4  Position    	: POSITION;
  half3  Normal      	: NORMAL;
#ifdef VERTEX_COLOR
  FIXED4 VtxColor    	  : COLOR0;
#endif
  half2  Tex0        	: TEXCOORD0;
#ifdef TANGENT_BASIS
  half3  Tangent     	: TANGENT;
  half3  Binormal    	: BINORMAL;
#endif
};

struct vsOut
{
  float4 HPosition			: POSITION;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  FIXED4 VtxColor			: COLOR0;
#endif

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW
  FIXED4 FogColor    		: COLOR1;
 #else	
  FIXED FogColor    		  : FOG;
 #endif	
#endif

  half4 Tex0				: TEXCOORD0;
#if RT_PROFILE <= QUALITY_LOW
  FIXED4 Diffuse			: COLOR0;
#endif

#if RT_PROFILE > QUALITY_VERYLOW
  half3 WorldNormal		: TEXCOORD2;
  float4 WorldView		: TEXCOORD3;
  float4 WorldPos	    : TEXCOORD4;
#endif  
#if defined(TANGENT_BASIS) && RT_PROFILE > QUALITY_VERYLOW
  half3 WorldTangent			: TEXCOORD5;
  half3 WorldBinormal		 	: TEXCOORD6;
#endif  
};


half3 GetLDiffuse(const float3 vWorldPos, const float3 vWorldNormal)
{
  half3 vDiffuseAcc = half3(0,0,0);

  float4 WorldLightPos = LGetPosition(0);
  half3 vVertexSpaceLightPos = WorldLightPos.xyz - vWorldPos;
  half3 vDirToLight = normalize(vVertexSpaceLightPos);
  float fNdotL = dot(vDirToLight.xyz, vWorldNormal);
  vDiffuseAcc.rgb = saturate(fNdotL) * LGetDiffuse(0);

  return vDiffuseAcc;
}

void LLighting(inout vsOut Out, const in vsIn In, const float3 vWorldPos, const float3 vWorldNormal)
{
#if RT_PROFILE <= QUALITY_LOW    
  Out.Diffuse = float4(0,0,0,1);
  #ifdef RT_NUM_LIGHTS
    Out.Diffuse.rgb += GetLDiffuse(vWorldPos, vWorldNormal.xyz);
  #endif
  #if defined(VERTEX_COLOR)
   #if !defined(RT_STATIC_LIGHTING)
    Out.Diffuse.rgb *= In.VtxColor.rgb;
   #endif	
    Out.Diffuse.a = In.VtxColor.a;
  #endif
  
  #ifndef RT_LIGHTMAP
    Out.Diffuse.rgb += F_WorldAmbient.rgb;
  #endif
#endif
}

vsOut CarVS(in vsIn In)
{
  vsOut Out = (vsOut)0;

  float4 Po = float4(In.Position.xyz, 1);
  float3 vWorldPos = MUL_VEC_MAT(Po, PI_mW).xyz;
  float3 vWorldNormal = normalize(MUL_VEC_MAT(In.Normal.xyz, (float3x3)PI_mW).xyz);

#if defined(TANGENT_BASIS)
  float3 vWorldTangent = normalize(MUL_VEC_MAT(In.Tangent.xyz, (float3x3)PI_mW).xyz);
  float3 vWorldBinormal = normalize(MUL_VEC_MAT(In.Binormal.xyz, (float3x3)PI_mW).xyz);
#endif
   
  Out.Tex0.xy = In.Tex0.xy;

#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  Out.VtxColor = In.VtxColor;
#endif

#if defined(RT_LIGHTMAP) || defined(SECONDARY_DIFFUSE)
  half fFracX = (LightmapTRCoords[1].x - vWorldPos.x) / (LightmapTRCoords[1].x - LightmapTRCoords[0].x);
  half fFracY = (vWorldPos.z - LightmapTRCoords[0].y) / (LightmapTRCoords[3].y - LightmapTRCoords[0].y);
  Out.Tex0.zw = lerp(lerp(LightmapTRCoords[0].zw, LightmapTRCoords[1].zw, fFracX), lerp(LightmapTRCoords[2].zw, LightmapTRCoords[3].zw, fFracX), fFracY);
#endif

//#if RT_PROFILE > QUALITY_VERYLOW
  Out.WorldNormal.xyz = vWorldNormal;  
 #if defined(TANGENT_BASIS)
  Out.WorldTangent = vWorldTangent;
  Out.WorldBinormal = vWorldBinormal;
 #endif
  Out.WorldView.xyz = MAT_POS(C_VI) - vWorldPos;
  Out.WorldPos.xyz = vWorldPos;
//#endif

  float4 vWPos = float4(vWorldPos, 1);
  Out.HPosition = MUL_VEC_MAT(vWPos, C_VP);

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW && !defined(OPENGL)
  Out.FogColor = Fog(length(Out.WorldView.xyz));
 #else
  Out.FogColor = Fog(Out.HPosition.w);
 #endif	
#endif

#if RT_PROFILE <= QUALITY_LOW
  //LLighting(Out, In, vWorldPos, vWorldNormal);
#endif

  return Out;
}

half3 rotateY(half3 v, float angle)
{
	float s, c;
	sincos(angle, s, c);
	return half3(c*v.x + s*v.z, v.y, -s*v.x + c*v.z);
}

FIXED4 CarPS(in vsOut IN) : COLOR
{
  half3 vNormal = IN.WorldNormal.xyz;
  half2 baseTC = IN.Tex0.xy;
  half2 lmTC = IN.Tex0.zw;
  half2 bumpTC = baseTC;

  half3 mTangentToWS0, mTangentToWS1, mTangentToWS2;
#if defined (TANGENT_BASIS)
  mTangentToWS0 = IN.WorldTangent.xyz;  
  mTangentToWS1 = IN.WorldBinormal.xyz;  
  mTangentToWS2 = IN.WorldNormal.xyz;  
#endif

  FIXED4 cDiffuseMap = tex2D(ColorSampler, baseTC);
  cDiffuseMap.rgb = cDiffuseMap.rgb + cDiffuseMap.a * CarColor.rgb;
  
  FIXED fAlpha = cDiffuseMap.a;
  FIXED4 cVertexColor = 1;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  cVertexColor = IN.VtxColor;
  fAlpha *= cVertexColor.a;
#endif

#if defined(SECONDARY_DIFFUSE) && defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  half4 vSecondColor = tex2D(SecondColorSampler, lmTC);
  float fA = cVertexColor.a;
  cDiffuseMap = lerp(cDiffuseMap, vSecondColor, fA);
#endif

  half fNormalA = 1.0f;
#if defined(BUMP_DIFFUSE) && defined(TANGENT_BASIS)
    half4 cBumpData = tex2D(SpecularSampler, bumpTC);
    fNormalA = cBumpData.a;
    half3 cBumpMap = half3(0, 0, 1); //cBumpData.rgb * 2.h - 1.h;
  #if RT_PROFILE > QUALITY_HIGH
    cBumpMap = normalize(cBumpMap);
  #endif
    vNormal.xyz  = cBumpMap.x * mTangentToWS0;
    vNormal.xyz += cBumpMap.y * mTangentToWS1;
    vNormal.xyz += cBumpMap.z * mTangentToWS2;
#endif
 

  float3 vView = normalize(IN.WorldView.xyz);
  float fNdotE = dot(vView, vNormal);
  half3 vReflVec = 2 * fNdotE * vNormal - vView;
  float fAngle = F_Time.x * ScrollSpeed;
  vReflVec = rotateY(vReflVec, -fAngle); // frac(IN.WorldPos.xyz * ReflDir) * 2;

  FIXED3 cDiffuseAcc = 0;
  FIXED3 cAmbientAcc = 0;
  FIXED3 cSpecularAcc = 0;
  FIXED fSpecularMap = 1;

  FIXED3 cLM = 1;
#if defined(RT_LIGHTMAP)
  cLM = tex2D(LightMapSampler, lmTC.xy).rgb * 2.h;
#endif

//#if RT_PROFILE > QUALITY_LOW
  cAmbientAcc = F_WorldAmbient.rgb;
//#endif
#if defined(RT_LIGHTMAP)
  cAmbientAcc = cLM;
#endif

//#if RT_PROFILE <= QUALITY_LOW
//  cDiffuseAcc = IN.Diffuse.rgb;
//#else

#if defined(RT_NUM_LIGHTS)

 #if defined(RT_SPECULAR)
  half fSpecMask = fNormalA;
  fSpecularMap *= fSpecMask;
 #endif

  // Assume this is sun
  FIXED3 cDiffuse = LGetDiffuse(0);
  FIXED3 cSpecular = LGetSpecular(0);

  half3 vDirToLight;
  FIXED fFallOff;
  GetLighting(vDirToLight, fFallOff, IN.WorldPos.xyz, LIGHT_DIRECTIONAL, 0, 0);
  half fNdotL = dot(vDirToLight, vNormal);
  cDiffuseAcc += cDiffuse * fNdotL;

 #if defined(RT_SPECULAR)
  half fSpec = Phong(vNormal, vView, vDirToLight, 8);
  cSpecular *= fSpec;
  cSpecularAcc += cSpecular * fSpecularMap * Ks * cLM;
 #endif
#endif

#if defined (RT_ENVMAP) && RT_PROFILE > QUALITY_VERYLOW
 #if defined (RT_TEX_ANIMATION)
  FIXED3 vEnv = tex2D(SpecularSampler, float2(1.0f-vReflVec.z, vReflVec.x)).rgb;
 #else
  FIXED3 vEnv = texCUBE(EnvMapSampler, vReflVec).rgb;
  //half fFresnelTerm = GetFresnel(fNdotE, RT_PROFILE, true);
 #endif 
  FIXED3 cEnv = vEnv * fSpecularMap * FresnelColor.xyz;
  cSpecularAcc += cEnv;
#endif

  FIXED4 cOut;
  cOut.xyz = (cAmbientAcc.xyz + cDiffuseAcc.xyz) * cDiffuseMap.xyz * cVertexColor.rgb + cSpecularAcc;
  cOut.w = fAlpha;

#if defined(RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  frag_fog_setup(IN.FogColor, cOut);
#endif

#if defined (RT_ENVMAP) && RT_PROFILE > QUALITY_VERYLOW
 #if defined (RT_TEX_ANIMATION)
  //cOut.xyz =  frac(IN.TexRefl.xyx);
 #endif
#endif

  return cOut;
}

technique Default
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE CarVS();
    PixelShader = compile PS_SHADER_TYPE CarPS();
    DEFAULT_STATES
  }
}
