#include "constants.fxh"
#include "lighting.fxh"

uniform float ScrollSpeed = 0.2f;

struct vsIn
{
  float4  Position    	: POSITION;
  half3  Normal      	: NORMAL;
  half2  Tex0        	: TEXCOORD0;
#ifdef TANGENT_BASIS
  half3  Tangent     	: TANGENT;
  half3  Binormal    	: BINORMAL;
#endif
};

struct vsOut
{
  float4 HPosition			: POSITION;

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW
  FIXED4 FogColor    		: COLOR1;
 #else	
  FIXED FogColor    		  : FOG;
 #endif	
#endif

  half4 Tex0				: TEXCOORD0;

#if RT_PROFILE > QUALITY_VERYLOW
  half3 WorldNormal		: TEXCOORD2;
  float4 WorldView		: TEXCOORD3;
  float4 WorldPos	    : TEXCOORD4;
 #if defined(TANGENT_BASIS)
  half3 WorldTangent			: TEXCOORD5;
  half3 WorldBinormal		 	: TEXCOORD6;
 #endif  
#endif  
};


vsOut CarVS(in vsIn In)
{
  vsOut Out = (vsOut)0;

  float4 Po = float4(In.Position.xyz, 1);
  float3 vWorldPos = MUL_VEC_MAT(Po, PI_mW).xyz;
  float3 vWorldNormal = normalize(MUL_VEC_MAT(In.Normal.xyz, (float3x3)PI_mW).xyz);

#if defined(TANGENT_BASIS)
  float3 vWorldTangent = normalize(MUL_VEC_MAT(In.Tangent.xyz, (float3x3)PI_mW).xyz);
  float3 vWorldBinormal = normalize(MUL_VEC_MAT(In.Binormal.xyz, (float3x3)PI_mW).xyz);
#endif
   
  Out.Tex0.xy = In.Tex0.xy;

#if defined(RT_LIGHTMAP)
  half fFracX = (LightmapTRCoords[1].x - vWorldPos.x) / (LightmapTRCoords[1].x - LightmapTRCoords[0].x);
  half fFracY = (vWorldPos.z - LightmapTRCoords[0].y) / (LightmapTRCoords[3].y - LightmapTRCoords[0].y);
  Out.Tex0.zw = lerp(lerp(LightmapTRCoords[0].zw, LightmapTRCoords[1].zw, fFracX), lerp(LightmapTRCoords[2].zw, LightmapTRCoords[3].zw, fFracX), fFracY);
#endif

  Out.WorldNormal.xyz = vWorldNormal;  
 #if defined(TANGENT_BASIS)
  Out.WorldTangent = vWorldTangent;
  Out.WorldBinormal = vWorldBinormal;
 #endif
  Out.WorldView.xyz = MAT_POS(C_VI) - vWorldPos;
  Out.WorldPos.xyz = vWorldPos;

  float4 vWPos = float4(vWorldPos, 1);
  Out.HPosition = MUL_VEC_MAT(vWPos, C_VP);

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW && !defined(OPENGL)
  Out.FogColor = Fog(length(Out.WorldView.xyz));
 #else
  Out.FogColor = Fog(Out.HPosition.w);
 #endif	
#endif

  return Out;
}

half3 rotateY(half3 v, float angle)
{
	float s, c;
	sincos(angle, s, c);
	return half3(c*v.x + s*v.z, v.y, -s*v.x + c*v.z);
}

FIXED4 CarPS(in vsOut IN) : COLOR
{
  half3 vNormal = IN.WorldNormal.xyz;
  half2 baseTC = IN.Tex0.xy;
  half2 lmTC = IN.Tex0.zw;
  half2 bumpTC = baseTC;

  half3 mTangentToWS0, mTangentToWS1, mTangentToWS2;
#if defined (TANGENT_BASIS)
  mTangentToWS0 = IN.WorldTangent.xyz;  
  mTangentToWS1 = IN.WorldBinormal.xyz;  
  mTangentToWS2 = IN.WorldNormal.xyz;  
#endif

  FIXED4 cDiffuseMap = tex2D(ColorSampler, baseTC);  
  FIXED fAlpha = cDiffuseMap.a;

#if defined(BUMP_DIFFUSE) && defined(TANGENT_BASIS)
    half4 cBumpData = tex2D(SpecularSampler, bumpTC);
    half3 cBumpMap = half3(0, 0, 1); //cBumpData.rgb * 2.h - 1.h;
  #if RT_PROFILE > QUALITY_HIGH
    cBumpMap = normalize(cBumpMap);
  #endif
    vNormal.xyz  = cBumpMap.x * mTangentToWS0;
    vNormal.xyz += cBumpMap.y * mTangentToWS1;
    vNormal.xyz += cBumpMap.z * mTangentToWS2;
#endif
 

  float3 vView = normalize(IN.WorldView.xyz);
  float fNdotE = dot(vView, vNormal);
  half3 vReflVec = 2 * fNdotE * vNormal - vView;
  float fShift = F_Time.x * ScrollSpeed;
  vReflVec = rotateY(vReflVec, -fShift); // frac(IN.WorldPos.xyz * ReflDir) * 2;

  FIXED3 cDiffuseAcc = 0;
  FIXED3 cAmbientAcc = 0;
  FIXED3 cSpecularAcc = 0;
  FIXED fSpecularMap = 1;

  FIXED3 cLM = 1;
#if defined(RT_LIGHTMAP)
  cLM = tex2D(LightMapSampler, lmTC.xy).rgb * 2.h;
#endif

  cAmbientAcc = F_WorldAmbient.rgb;
#if defined(RT_LIGHTMAP)
  cAmbientAcc = cLM;
#endif

#if defined(RT_NUM_LIGHTS)

 #if defined(RT_SPECULAR)
  fSpecularMap = fAlpha;
 #endif

  // Assume this is sun
  FIXED3 cDiffuse = LGetDiffuse(0);
  FIXED3 cSpecular = LGetSpecular(0);

  half3 vDirToLight;
  FIXED fFallOff;
  GetLighting(vDirToLight, fFallOff, IN.WorldPos.xyz, LIGHT_DIRECTIONAL, 0, 0);
  half fNdotL = dot(vDirToLight, vNormal);
  cDiffuseAcc += cDiffuse * fNdotL;

 #if defined(RT_SPECULAR)
  half fSpec = Phong(vNormal, vView, vDirToLight, 8);
  cSpecular *= fSpec;
  cSpecularAcc += cSpecular * fSpecularMap * Ks * cLM;
 #endif
#endif

#if defined (RT_ENVMAP) && RT_PROFILE > QUALITY_VERYLOW
  FIXED3 vEnv = texCUBE(EnvMapSampler, vReflVec).rgb;
  half fLerp = abs(vNormal.z);
  fLerp *= fLerp;
 #if defined(SECONDARY_DIFFUSE)
  half2 shiftTC = baseTC;
  shiftTC.x += fShift;
  half4 vSideColor = tex2D(SecondColorSampler, shiftTC);
  vEnv = lerp(vEnv, vSideColor, fLerp);
 #else
  vEnv *= (1.0-fLerp); 
 #endif  
  half fFresnelTerm = 1.0f;
 #if defined (RT_FRESNEL) && RT_PROFILE > QUALITY_VERYLOW
  fFresnelTerm = GetFresnel(fNdotE, RT_PROFILE, true);
 #endif 
  FIXED3 cEnv = vEnv * fSpecularMap * fFresnelTerm * Ks * FresnelColor.xyz;
  cSpecularAcc += cEnv;
#endif

  FIXED4 cOut;
  cOut.xyz = (cAmbientAcc.xyz + cDiffuseAcc.xyz) * cDiffuseMap.xyz + cSpecularAcc;
  cOut.w = fAlpha;

#if defined(RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  frag_fog_setup(IN.FogColor, cOut);
#endif

  return cOut;
}

technique Default
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE CarVS();
    PixelShader = compile PS_SHADER_TYPE CarPS();
    DEFAULT_STATES
  }
}
