/*
	Depth of field
*/

#include "constants.fxh"

sampler2D depthMap = sampler_state
{
    Texture = <Texture0>;
    MinFilter = POINT;
    MagFilter = POINT;
    MipFilter = NONE;
    AddressU = Clamp;
    AddressV = Clamp;
};
sampler2D texMap = sampler_state
{
    Texture = <Texture0>;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
    MipFilter = NONE;
    AddressU = Clamp;
    AddressV = Clamp;
};
sampler2D backMap = sampler_state
{
    Texture = <Texture0>;
    MinFilter = POINT;
    MagFilter = POINT;
    MipFilter = NONE;
    AddressU = Clamp;
    AddressV = Clamp;
};


extern const float4 ParamsFocus;
extern const float4 ParamsBlur;
extern const float4 ParamsPixelSizes;

struct VS_INPUT 
{
   float3 Pos : POSITION;
   float2 tc0 : TEXCOORD0;
};

struct vtxDOF
{
  float4 HPosition  : POSITION;
  float2 tc0        : TEXCOORD0;
};

vtxDOF DofVS(VS_INPUT IN)
{
  vtxDOF OUT = (vtxDOF)0; 

	OUT.HPosition = float4(IN.Pos, 1);
  OUT.HPosition.xy *= F_ScreenSize.xy;
  OUT.HPosition.xy = OUT.HPosition.xy * 2.f - 1.f;

  OUT.tc0.xy = IN.tc0.xy;  
    
  return OUT;
}
vtxDOF DofFinalVS(VS_INPUT IN)
{
  vtxDOF OUT = (vtxDOF)0; 

	OUT.HPosition = float4(IN.Pos, 1);
  OUT.HPosition.xy *= F_ScreenSize.xy;
  OUT.HPosition.xy = OUT.HPosition.xy * 2.f - 1.f;

  OUT.tc0.xy = IN.tc0.xy;  
#ifdef OPENGL
  OUT.tc0.y = 1 - OUT.tc0.y;
#endif
    
  return OUT;
}

half4 DOF_Pass_PS(vtxDOF IN) : COLOR0
{
  half4 OUTColor = 0;

  const int tapCount = 8;

  const half4 dofParams = half4(0.2, 0.4, 0, 0);

  float2 poisson[8] =
  {
     0.0,    0.0,
     0.527, -0.085,
    -0.040,  0.536,
    -0.670, -0.179,
    -0.419, -0.616,
     0.440, -0.639,
    -0.757,  0.349,
     0.574,  0.685
  };

  half4 cOut = 0;
  half discRadius;
  half discRadiusLow;
  half centerDepth;
	half centerDepthLow;
        
	centerDepth = tex2D(backMap, IN.tc0.xy).w; 
  centerDepthLow = tex2D(texMap, IN.tc0.xy).w; 

  discRadius = (centerDepth * ParamsBlur.y - ParamsBlur.x);
  discRadiusLow = discRadius * dofParams.y;
  
  half4 pixelSizes;
  pixelSizes.xy = ParamsPixelSizes.xy * discRadius;
  pixelSizes.wz = ParamsPixelSizes.zw * discRadiusLow;

  for (int t=0; t<tapCount; t++)
  { 
    half4 tapHigh = tex2D(backMap, IN.tc0.xy + poisson[t] * pixelSizes.xy);                
    half4 tapLow = tex2D(texMap, IN.tc0.xy + poisson[t] * pixelSizes.wz);        
        
    half tapLerp = tapHigh.a * 2.0 - 1.0;        
    half4 tap = lerp(tapHigh, tapLow, saturate(tapLerp));    
    
    tap.a = (tapLow.a - centerDepth + dofParams.x > 0.0) ? 1 : saturate(tap.a * 2.0 - 1.0);    
        
    cOut.xyz += tap.a * tap.xyz;
    cOut.w += tap.a;
  }

  OUTColor = cOut / cOut.w;
	
  return OUTColor;
}

half DepthBlurinessBiased(half fDepth)
{
  half f = 0; 

  if (fDepth > ParamsFocus.y)
  {
		f = saturate((fDepth - ParamsFocus.y) / ParamsFocus.z);
  }
  else
  if (fDepth < ParamsFocus.x)
  {
    f = saturate((1 - fDepth/ParamsFocus.x) / ParamsFocus.w);
		f *= -1;
  }

  return f;
}

half4 DepthToAlphaPS(vtxDOF IN) : COLOR0
{
  half4 OUT;  
    
  half4 depthNormalized = tex2D(depthMap, IN.tc0.xy) * C_FarClip.x;
  half depth = DepthBlurinessBiased(depthNormalized) * ParamsFocus.w;	  
    
  OUT = depth * 0.5h + 0.5h;
  
  return OUT;
}

technique Main
{
	pass p0 
	{		
		VertexShader = compile VS_SHADER_TYPE DofFinalVS();
		PixelShader  = compile PS_SHADER_TYPE DOF_Pass_PS();
#ifdef OGL
		CullMode = FrontAndBack;
#else
		CullMode = none;
#endif
		ZEnable = false;
	}
}

technique DepthToAlpha
{
  pass p0
  {        
#ifdef OGL
		CullMode = FrontAndBack;
    ColorMask = { 1, 0, 0, 0 };
#else
		CullMode = none;
    ColorWriteEnable = Alpha;
#endif
            
    VertexShader = compile VS_SHADER_TYPE DofVS();
    PixelShader = compile PS_SHADER_TYPE DepthToAlphaPS();    

  }
}


half4 ParamOffsets[16];
half4 ParamWeights[16];

struct vtxBlur
{
  float4 HPosition : POSITION;
  float4 tc0 : TEXCOORD0;    
  float4 tc1 : TEXCOORD1;    
  float4 tc2 : TEXCOORD2;    
  float4 tc3 : TEXCOORD3;    
};

vtxBlur BlurVS(VS_INPUT IN)
{
  vtxBlur OUT = (vtxBlur)0;

	OUT.HPosition = float4(IN.Pos, 1);
  OUT.HPosition.xy *= F_ScreenSize.xy;
  OUT.HPosition.xy = OUT.HPosition.xy * 2.f - 1.f;
  
  OUT.tc0.xy = IN.tc0.xy + ParamOffsets[0].xy;
  OUT.tc0.zw = IN.tc0.xy + ParamOffsets[1].xy;
  OUT.tc1.xy = IN.tc0.xy + ParamOffsets[2].xy;
  OUT.tc1.zw = IN.tc0.xy + ParamOffsets[3].xy;
  OUT.tc2.xy = IN.tc0.xy + ParamOffsets[4].xy;
  OUT.tc2.zw = IN.tc0.xy + ParamOffsets[5].xy;
  OUT.tc3.xy = IN.tc0.xy + ParamOffsets[6].xy;
  OUT.tc3.zw = IN.tc0.xy + ParamOffsets[7].xy;

  return OUT;
}

half4 BlurPS(vtxBlur IN) : COLOR0
{
  half4 sum = 0;
    
	half4 col = tex2D(texMap, IN.tc0.xy);
	sum += col * ParamWeights[0].x;  

	col = tex2D(texMap, IN.tc0.zw);
	sum += col * ParamWeights[1].x;  
	
  col = tex2D(texMap, IN.tc1.xy);
	sum += col * ParamWeights[2].x;  

	col = tex2D(texMap, IN.tc1.zw);
	sum += col * ParamWeights[3].x;

	col = tex2D(texMap, IN.tc2.xy);
	sum += col * ParamWeights[4].x;  
	
	col = tex2D(texMap, IN.tc2.zw);
	sum += col * ParamWeights[5].x;  
	
	col = tex2D(texMap, IN.tc3.xy);
	sum += col * ParamWeights[6].x;  
	
	col = tex2D(texMap, IN.tc3.zw);
	sum += col * ParamWeights[7].x;

  return sum;
}

// Optimized gauss blur version, making use of bilinear filtering
technique Blur
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE BlurVS();
    PixelShader = compile PS_SHADER_TYPE BlurPS();    
  }
}


//===========================================================================================

struct vs_in {
	float3 Pos: POSITION;
	float2 Tex0: TEXCOORD0;
};

struct vs_out {
	float4 Pos: POSITION;
	float2  Tex0: TEXCOORD0;
};

sampler2D samplerTex = sampler_state
{
    Texture = <Texture0>;
    MinFilter = Point;
    MagFilter = Point;
    MipFilter = None;
    AddressU = Clamp;
    AddressV = Clamp;
};

vs_out VSShow(vs_in In)
{
	vs_out Out;
	Out.Pos = float4(In.Pos.xyz, 1);
  Out.Pos.xy *= F_ScreenSize.xy;
  Out.Pos.xy = Out.Pos.xy * 2.f - 1.f;
	Out.Tex0 = In.Tex0;
	return Out;
}

float4 PSShow(vs_out In): COLOR
{
	float4 Col = tex2D(samplerTex, In.Tex0);
	
	return Col;
}

technique TShow
{
	pass p0 
	{
		VertexShader = compile VS_SHADER_TYPE VSShow();
		PixelShader  = compile PS_SHADER_TYPE PSShow();
	}
}
