#include "transform.fxh"

//============================================================================================================================

struct v2f
{
  float4 hPos			: POSITION; 
  float3 hPosTexProj	: TEXCOORD0;
  float3 viewDir		  : TEXCOORD1;
  float3 cameraDir		: TEXCOORD2;
  float4 fogPlane     : TEXCOORD3;
};
struct v2fl
{
  float4 hPos			: POSITION; 
  float2 texFog	  : TEXCOORD0;
  float2 texFogEnter : TEXCOORD1;
};


sampler2D samplerSceneZ = sampler_state
{
    Texture = <Texture0>;
    MinFilter = Point;
    MagFilter = Point;
    MipFilter = None;
    AddressU = Clamp;
    AddressV = Clamp;
};
float4 WaterFogColor : register(c10);

float4 cFogColorShallowWaterLevel;
float cUnderWaterInScatterConst;
float4 FogPlane;
float4 ViewerNear;


float GetSceneDepthInto(in float3 screenProj)
{
	float depth = tex2D(samplerSceneZ, screenProj.xy/screenProj.z).r * C_FarClip.x;
	return depth;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Case: Looking into fog volume
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


v2f FogIntoVS(in VS_IN IN)
{
  v2f OUT = (v2f)0; 

  float3 WorldPos = WorldPosition(IN);
  float4 hPos = mul(float4(WorldPos,1),C_VP);

  OUT.hPos = hPos;
  OUT.hPos.z *= ViewerNear.x; // effectively prevent any clipping and depth test rejection by muliplying homogeneous z by zero (if close enough to water surface)

  OUT.hPosTexProj = HPosToScreenTC(hPos);
    
  OUT.viewDir = WorldPos - C_Eye;
  OUT.cameraDir.xyz = C_VI[2].xyz;
  
  return OUT;
}


half4 FogIntoPS(v2f IN) : COLOR
{
  float fFogDensity = 1.44269502f * WaterFogColor.w; // log2(e) = 1.44269502
  float sceneDepth = GetSceneDepthInto(IN.hPosTexProj);
  float volumeDepth = max(sceneDepth - dot(IN.viewDir, IN.cameraDir), 0);  
  float waterVolumeFog = exp2(-fFogDensity * volumeDepth / dot(normalize(IN.viewDir), IN.cameraDir));
  
  half4 Color = half4(WaterFogColor.xyz, waterVolumeFog);

  return Color;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Case: Looking out of water volume
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct VS_INPUT4 
{
   float4 Pos : POSITION;
   float3 tc0 : TEXCOORD0;
};

v2f FogOutofVS(in VS_INPUT4 IN)
{
  v2f OUT = (v2f) 0; 

  OUT.hPos = IN.Pos;
  OUT.hPos.xy *= F_ScreenSize.xy;
  OUT.hPos.xy = OUT.hPos.xy * 2.f - 1.f;

  OUT.hPos.z = 0; // always prevent any clipping and depth test rejection as we draw full screen quad
	
  OUT.hPosTexProj = HPosToScreenTC(OUT.hPos);
    
  OUT.viewDir = IN.tc0.xyz;
  OUT.cameraDir = C_VI[2].xyz;

  return OUT;
}

half4 FogOutofPS( v2f IN ) : COLOR
{
  float3 viewDir = IN.viewDir;		
		
  float fFogDensity = 1.44269502f * WaterFogColor.w; // log2(e) = 1.44269502
  float sceneDepth = GetSceneDepthInto(IN.hPosTexProj);
  float volumeDepth = min(dot(viewDir, IN.cameraDir), sceneDepth);  
  float waterVolumeFog = exp2(-fFogDensity * volumeDepth / dot(normalize(viewDir), IN.cameraDir));
  
  half4 Color = half4(WaterFogColor.xyz, waterVolumeFog);

  return Color;
}

//=============================================================================================================================

technique FogInto
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE FogIntoVS();
    PixelShader = compile PS_SHADER_TYPE FogIntoPS();
    ZEnable = false;
    AlphaBlendEnable = true;
#ifdef OGL
		CullMode = FrontAndBack;
    BlendFunc = { SrcAlpha, InvSrcAlpha };
#else
    CullMode = None;
    SrcBlend = InvSrcAlpha;
    DestBlend = SrcAlpha;
#endif    
    DEFAULT_STATES
  }
}

technique FogOutof
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE FogOutofVS();
    PixelShader = compile PS_SHADER_TYPE FogOutofPS();
    ZEnable = false;
    AlphaBlendEnable = true;
#ifdef OGL
		CullMode = FrontAndBack;
    BlendFunc = { SrcAlpha, InvSrcAlpha };
#else
    CullMode = None;
    SrcBlend = InvSrcAlpha;
    DestBlend = SrcAlpha;
#endif    
    DEFAULT_STATES
  }
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Case: Looking into fog volume
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

v2f OceanIntoVS(in VS_IN IN)
{
  v2f OUT = (v2f)0; 

  float3 WorldPos = IN.Position.xyz;
  float4 hPos = mul(float4(WorldPos, 1), C_VP);

  OUT.hPos = hPos;
  OUT.hPos.z *= ViewerNear.x; // effectively prevent any clipping and depth test rejection by muliplying homogeneous z by zero (if close enough to water surface)

  OUT.hPosTexProj = HPosToScreenTC(hPos);
    
  OUT.viewDir = WorldPos - C_Eye;
  OUT.cameraDir.xyz = C_VI[2].xyz;
  
  return OUT;
}


half4 OceanIntoPS(v2f IN) : COLOR
{
  float fFogDensity = /*1.44269502f * */WaterFogColor.w; // log2(e) = 1.44269502
  float sceneDepth = GetSceneDepthInto(IN.hPosTexProj);
  float volumeDepth = max(sceneDepth - dot(IN.viewDir, IN.cameraDir), 0);  
  float waterVolumeFog = exp2(-fFogDensity * volumeDepth / dot(normalize(IN.viewDir), IN.cameraDir));
  
  half4 Color = half4(WaterFogColor.xyz, waterVolumeFog);

  return Color;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Case: Looking out of water volume
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

v2f OceanOutofVS(in VS_INPUT4 IN)
{
  v2f OUT = (v2f) 0; 

  OUT.hPos = IN.Pos;
  OUT.hPos.xy *= F_ScreenSize.xy;
  OUT.hPos.xy = OUT.hPos.xy * 2.f - 1.f;

  OUT.hPos.z = 0; // always prevent any clipping and depth test rejection as we draw full screen quad
	
  OUT.hPosTexProj = HPosToScreenTC(OUT.hPos);
    
  OUT.viewDir = IN.tc0.xyz;
  OUT.cameraDir = C_VI[2].xyz;
  OUT.fogPlane = FogPlane;
  float fPerpdist = dot(C_Eye, OUT.fogPlane.xyz) + OUT.fogPlane.w;
  OUT.fogPlane.w = fPerpdist;

  return OUT;
}

half4 OceanOutofPS (v2f IN) : COLOR
{
  float3 viewDir = IN.viewDir;				
  
  float fFogDensity = 1.44269502f * WaterFogColor.w; // log2(e) = 1.44269502
  
	float fCosine = dot(viewDir, IN.fogPlane.xyz);
	float pd_c = -IN.fogPlane.w / fCosine;			
	float t = IN.fogPlane.w + dot(viewDir, IN.fogPlane.xyz);	
	if (t > 0)
		viewDir *= pd_c;
  
  float sceneDepth = GetSceneDepthInto(IN.hPosTexProj);
  float volumeDepth = min(dot(viewDir, IN.cameraDir), sceneDepth);  
  float waterVolumeFog = exp2(-fFogDensity * volumeDepth / dot(normalize(viewDir), IN.cameraDir));
  
  half4 Color = half4(WaterFogColor.xyz, waterVolumeFog);

  return Color;
}

technique OceanInto
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE OceanIntoVS();
    PixelShader = compile PS_SHADER_TYPE OceanIntoPS();
    ZEnable = false;
    AlphaBlendEnable = true;
#ifdef OGL
		CullMode = FrontAndBack;
    BlendFunc = { SrcAlpha, InvSrcAlpha };
#else
    CullMode = None;
    SrcBlend = InvSrcAlpha;
    DestBlend = SrcAlpha;
#endif    
    DEFAULT_STATES
  }
}

technique OceanOutof
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE OceanOutofVS();
    PixelShader = compile PS_SHADER_TYPE OceanOutofPS();
    ZEnable = false;
    AlphaBlendEnable = true;
#ifdef OGL
		CullMode = FrontAndBack;
    BlendFunc = { SrcAlpha, InvSrcAlpha };
#else
    CullMode = None;
    SrcBlend = InvSrcAlpha;
    DestBlend = SrcAlpha;
#endif    
    DEFAULT_STATES
  }
}


v2fl OceanLowVS(in VS_IN IN)
{
  v2fl OUT = (v2fl)0; 

  float3 WorldPos = WorldPosition(IN);

  float4 hPos = mul(float4(WorldPos, 1), C_VP);
  OUT.hPos = hPos;
    
#if RT_PROFILE == QUALITY_LOW
  OUT.texFog.x = dot(C_FogMatrix[0], IN.Position);
  OUT.texFog.y = dot(C_FogMatrix[1], IN.Position);
  OUT.texFogEnter.x = dot(C_FogEnterMatrix[0], IN.Position);
  OUT.texFogEnter.y = dot(C_FogEnterMatrix[1], IN.Position);
#endif
  
  return OUT;
}


half4 OceanLowPS(v2fl IN) : COLOR
{
  half4 fog = tex2D(samplerFog, IN.texFog);
  half4 fogEnter = tex2D(samplerFogEnter, IN.texFogEnter);
  half4 Color = fog * fogEnter;
  Color *= WaterFogColor;

  return Color;
}

technique OceanLow
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE OceanLowVS();
    PixelShader = compile PS_SHADER_TYPE OceanLowPS();
    ZEnable = false;
    AlphaBlendEnable = true;
#ifdef OGL
		CullMode = FrontAndBack;
    BlendFunc = { SrcAlpha, InvSrcAlpha };
#else
    CullMode = None;
    SrcBlend = InvSrcAlpha;
    DestBlend = SrcAlpha;
#endif    
    DEFAULT_STATES
  }
}

