#include "constants.fxh"
#include "lighting.fxh"

uniform float4 BumpOffsetScale = float4(0, 0, 1, 1);

struct vsIn
{
  float4  Position    	: POSITION;
  half3  Normal      	: NORMAL;
#ifdef VERTEX_COLOR
  FIXED4 VtxColor    	  : COLOR0;
#endif
  half2  Tex0        	: TEXCOORD0;
#if defined(RT_LIGHTMAP) && !defined(RT_CUSTOM_LIGHTMAP)
  half2  Tex1        	: TEXCOORD1;
#endif
};

struct vsOut
{
  float4 HPosition			: POSITION;
#if defined(VERTEX_COLOR)
  FIXED4 VtxColor			: COLOR0;
#endif

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW
  FIXED4 FogColor    		: COLOR1;
 #else	
  FIXED FogColor    		  : FOG;
 #endif	
#endif

  float4 Tex0				: TEXCOORD0;
#if defined(RT_LIGHTMAP) || defined(SECONDARY_DIFFUSE)
  half2  Tex1      	: TEXCOORD1;
#endif

#if RT_PROFILE > QUALITY_VERYLOW
  half3 WorldNormal			: TEXCOORD2;
  float4 WorldView			: TEXCOORD3;
  float4 WorldPos			  : TEXCOORD4;
#endif  
};


vsOut RoadVS(in vsIn In)
{
  vsOut Out = (vsOut)0;

  float4 Po = float4(In.Position.xyz, 1);
  float3 vWorldPos = MUL_VEC_MAT(Po, PI_mW).xyz;
  float3 vWorldNormal = normalize(MUL_VEC_MAT(In.Normal.xyz, (float3x3)PI_mW).xyz);

  Out.Tex0.xy = In.Tex0.xy;
  Out.Tex0.zw = (In.Tex0.xy + BumpOffsetScale.xy) * BumpOffsetScale.zw;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  Out.VtxColor = In.VtxColor;
#endif

#if defined(RT_LIGHTMAP) || defined(SECONDARY_DIFFUSE)
  #if defined(RT_CUSTOM_LIGHTMAP)
    Out.Tex1.xy = (vWorldPos.xz - LightmapCoords.xy) * LightmapCoords.z;
    Out.Tex1.xy += LightmapCoords.w;
    Out.Tex1.y = 1 - Out.Tex1.y;
  #elif defined(RT_LIGHTMAP)
    Out.Tex1.xy = In.Tex1.xy;
  #endif	
#endif

#if RT_PROFILE > QUALITY_VERYLOW
  Out.WorldNormal.xyz = vWorldNormal;  
  Out.WorldView.xyz = MAT_POS(C_VI) - vWorldPos;
  Out.WorldPos.xyz = vWorldPos;
#endif

  float4 vWPos = float4(vWorldPos, 1);
  Out.HPosition = MUL_VEC_MAT(vWPos, C_VP);

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW && !defined(OPENGL)
  Out.FogColor = Fog(length(Out.WorldView.xyz));
 #else
  Out.FogColor = Fog(Out.HPosition.w);
 #endif	
#endif

  return Out;
}

FIXED4 RoadPS(in vsOut IN) : COLOR
{
  half3 vNormal = IN.WorldNormal.xyz;
  float2 baseTC = IN.Tex0.xy;
  float2 specTC = IN.Tex0.zw;

#if defined(RT_LIGHTMAP) || defined(SECONDARY_DIFFUSE)
  half2 lmTC = IN.Tex1.xy;
#endif

  FIXED4 cDiffuseMap = tex2D(ColorSampler, baseTC);

  FIXED fAlpha = cDiffuseMap.a;
  FIXED4 cVertexColor = 1;

#if defined (VERTEX_COLOR)
  cVertexColor = IN.VtxColor;
  fAlpha *= cVertexColor.a;
#endif

  FIXED fSpecularMap = 1;

#if defined(SECONDARY_DIFFUSE) && defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  FIXED4 vSecondColor = tex2D(SecondColorSampler, lmTC);
  FIXED fA = cVertexColor.a;
 #if defined(ROAD_BLEND)
    fA *= cDiffuseMap.a;
    half blendFactor = fA * (1 + BlendParams.x);
    fA = saturate(pow(blendFactor, BlendParams.y));  // Falloff
 #endif
  cDiffuseMap = lerp(cDiffuseMap, vSecondColor, fA);
#endif

  FIXED3 cDiffuseAcc = 0;
  FIXED3 cAmbientAcc = F_WorldAmbient.rgb;
  #if defined(RT_SPECULAR)
   FIXED3 cSpecularAcc = 0;
  #endif

  FIXED3 cLM = 1;
#if defined(RT_LIGHTMAP)
  cLM = tex2D(LightMapSampler, lmTC.xy).rgb * 2.h;
  cAmbientAcc = cLM;
#endif

#if defined(RT_NUM_LIGHTS)

 #if defined(RT_SPECULAR)
  half fSpecMask = tex2D(ColorSampler, specTC).a;
  fSpecularMap *= fSpecMask;
  vNormal.xyz = IN.WorldNormal.xyz;
 #endif

  float3 vView = normalize(IN.WorldView.xyz);

  // Assume this is sun
  FIXED3 cDiffuse = LGetDiffuse(0);
  #if defined(RT_SPECULAR)
   FIXED3 cSpecular = LGetSpecular(0);
  #endif

  half3 vDirToLight;
  FIXED fFallOff;
  GetLighting(vDirToLight, fFallOff, IN.WorldPos.xyz, LIGHT_DIRECTIONAL, 0, 0);
  half fNdotL = dot(vDirToLight, vNormal);
  cDiffuseAcc += cDiffuse * fNdotL;

 #if defined(RT_SPECULAR)
  half fSpec = Phong(vNormal, vView, vDirToLight, 8);
  cSpecular *= fSpec;
  cSpecularAcc += cSpecular * fSpecularMap * cLM;
 #endif
#endif

  FIXED4 cOut;
  #if defined(RT_SPECULAR) && defined(RT_NUM_LIGHTS)
  cOut.xyz = (cAmbientAcc.xyz + cDiffuseAcc.xyz) * cDiffuseMap.xyz * cVertexColor.rgb * fSpecMask + cSpecularAcc;
  #else
  half fSpecMask = tex2D(ColorSampler, specTC).a;
  cOut.xyz = (cAmbientAcc.xyz + cDiffuseAcc.xyz) * cDiffuseMap.xyz * cVertexColor.rgb * fSpecMask;
  #endif
  cOut.w = fAlpha;

#if defined(RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  frag_fog_setup(IN.FogColor, cOut);
#endif

#if defined (RT_SPECULAR) && RT_PROFILE > QUALITY_VERYLOW
  //cOut.xyz = cSpecularAcc;
#endif

  return cOut;
}

technique Default
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE RoadVS();
    PixelShader = compile PS_SHADER_TYPE RoadPS();
    DEFAULT_STATES
  }
}
