#include "constants.fxh"

uniform float4 vShadowParams; // fShadowsFadeDist, fCascadeIndex, fShadowBias
uniform float4 vCEye; // fShadowsFadeDist, fCascadeIndex, fShadowBias

uniform float4x4 mShadowWLP0;
uniform float4x4 mShadowWL0;	// view matrix
uniform float4x4 mLightToCamera;

uniform float4 mSMSize;	// x - size, y - 1.f/size, zw - proj ratio

#define MAX_SHADOW_SAMPLES_NUM 8
uniform float4 vPoissonCoeffs[MAX_SHADOW_SAMPLES_NUM/2];

texture2D ZBuffer;
texture2D ShadowMapLOD;
texture2D JitterTexture;

sampler2D samplerSceneZ = sampler_state
{
  Texture = <ZBuffer>;
  MinFilter = Point;
  MagFilter = Point;
  MipFilter = None;
  AddressU = Border;
  AddressV = Border;
#ifdef OPENGL
  //BorderColor = { 1,1,1,1 };
#else	
  BorderColor = 0xffffffff;
#endif	
};

sampler2D samplerSMLOD = sampler_state
{
  Texture = <ShadowMapLOD>;
  MinFilter = Point;
  MagFilter = Point;
  MipFilter = None;
  AddressU = Clamp;
  AddressV = Clamp;
};

sampler2D sampler0 = sampler_state
{
  Texture = <Texture0>;
  MinFilter = Linear;
  MagFilter = Linear;
  MipFilter = None;
  AddressU = Clamp;
  AddressV = Clamp;
};
sampler2D sampler1 = sampler_state
{
  Texture = <Texture1>;
  MinFilter = Linear;
  MagFilter = Linear;
  MipFilter = None;
	AddressU = Clamp;
	AddressV = Clamp;
};

sampler2D JitterSampler = sampler_state
{
  Texture = <JitterTexture>;
  MinFilter = Point;
  MagFilter = Point;
  MipFilter = None;
  AddressU = Wrap;
  AddressV = Wrap;
};

half2 GetKernelSize(uniform int nNumSamples)
{
  return max(float2(1.0f, 1.0f), float2(1, 1) / pow(2, vShadowParams.y));
}

struct VS_INPUT_STENCIL
{
	float4 Pos				: POSITION;
	float2 Tex0				: TEXCOORD0;
};

struct VS_INPUT_FULLSCREEN
{
	float4 Pos				: POSITION;
	float2 Tex0				: TEXCOORD0;
	float3 CameraRay	: TEXCOORD1;
};

struct vertexOutput
{
  float4 HPosition       : POSITION;
  float3 Tex0_ShadowBias : TEXCOORD0;
  half3 CameraRay        : TEXCOORD1;
#if RT_PROFILE > QUALITY_MEDIUM
  half4 noisePosVec      : TEXCOORD2;
  half4 kernelSize       : TEXCOORD3;
#endif
};

struct stencilPrepassOutput
{
  float4 HPosition			: POSITION;
};

stencilPrepassOutput StencilPrepass_VS(in VS_INPUT_STENCIL In)
{
  stencilPrepassOutput Out;
#ifndef OPENGL
  Out.HPosition = MUL_VEC_MAT(float4(In.Pos.xy - float2(mSMSize.y * .5f, mSMSize.y * .5f), In.Pos.z, 1), mLightToCamera);
#else
  Out.HPosition = MUL_VEC_MAT(float4(In.Pos.xyz, 1), mLightToCamera);
#endif
  return Out;
}

vertexOutput _ScreenSpace_VS(in VS_INPUT_FULLSCREEN In, const int nNumSamples)
{
  vertexOutput Out;
  Out.HPosition = In.Pos;

  Out.Tex0_ShadowBias.xy = In.Tex0;
  Out.Tex0_ShadowBias.z = vShadowParams.z * (vShadowParams.y * .6f + 1.f);

  Out.CameraRay = In.CameraRay;

#if RT_PROFILE > QUALITY_MEDIUM
  Out.noisePosVec.xy = MUL_VEC_MAT(vCEye.xyz, (float3x3)mShadowWL0).xy;
  Out.noisePosVec.zw = MUL_VEC_MAT(In.CameraRay + vCEye.xyz, (float3x3)mShadowWL0).xy - Out.noisePosVec.xy;

  Out.kernelSize.xy = GetKernelSize(nNumSamples) * 15.h; // 5 for non-fs shadow pass
  Out.kernelSize.zw = GetKernelSize(nNumSamples) * mSMSize.y;
#endif

  return Out;
}

vertexOutput ScreenSpace_VSJit(in VS_INPUT_FULLSCREEN In)
{
  return _ScreenSpace_VS(In, MAX_SHADOW_SAMPLES_NUM);
}
vertexOutput ScreenSpace_VSHJit(in VS_INPUT_FULLSCREEN In)
{
  return _ScreenSpace_VS(In, MAX_SHADOW_SAMPLES_NUM/2);
}
vertexOutput ScreenSpace_VS0(in VS_INPUT_FULLSCREEN In)
{
  return _ScreenSpace_VS(In, 0);
}

half2 DoubleSampleRotated(in float2 shCoords, in half4 rotMatr, in half2 kernel1, const in half2 kernel2)
{
	// Rotate tap for this pixel location
	half4 rotatedOff =	rotMatr.xyzw * float4(kernel1.xx, kernel2.yy) +
											rotMatr.zwxy * float4(kernel1.yy, kernel2.xx);
	float4 fetchPos = shCoords.xyxy + rotatedOff;
	half2 result;
	result.x = tex2D( samplerSMLOD, fetchPos.xy ).r;
	result.y = tex2D( samplerSMLOD, fetchPos.zw ).r;
	return result;
}

half CalcShadowPoissonFilter(in float2 shCoords, in float depth, in float2 randDirTC, in half4 kernelSize, uniform int nNumSamples)
{
	// rotation 2x2 matrix for DoubleSampleRotated
	half2 rotSample = tex2D(JitterSampler, randDirTC.xy * kernelSize.xy).rg * kernelSize.zw;
	half4 rot = float4(rotSample.x, -rotSample.y, rotSample.y, rotSample.x);

	half4 shadowTest = 0;
	#define KERNEL_STEP_SIZE 2
	#define KERNEL_STEPS (nNumSamples / 2.h)
	for(int i=0; i<KERNEL_STEPS; i+=KERNEL_STEP_SIZE) // Loop over taps
	{
		half4 sampleDepth;
		sampleDepth.xy = DoubleSampleRotated(shCoords, rot, vPoissonCoeffs[i+0].xy, vPoissonCoeffs[i+0].zw);
		sampleDepth.zw = DoubleSampleRotated(shCoords, rot, vPoissonCoeffs[i+1].xy, vPoissonCoeffs[i+1].zw);

		// Determine whether tap is in shadow
    shadowTest += (sampleDepth > depth.xxxx);
	}

	#define INV_SAMPL_NUM (1.h / nNumSamples)
	return dot(shadowTest, INV_SAMPL_NUM);
}

half CalcShadowBilerp(in half2 smCoords, const in float depth)
{
  FIXED shadow[4] = {0, 0, 0, 0};
  if (tex2D(samplerSMLOD, smCoords).r >= depth)
    shadow[0] = 1.h;

  //half fS = tex2D(samplerSMLOD, smCoords).r;
  //return depth;

  if(GetShaderQuality() <= QUALITY_LOW)
    return shadow[0];

  if (tex2D(samplerSMLOD, smCoords + half2(1, 0) * mSMSize.y).r >= depth)
    shadow[1] = 1.h;
  if (tex2D(samplerSMLOD, smCoords + half2(0, 1) * mSMSize.y).r >= depth)
    shadow[2] = 1.h;
  if (tex2D(samplerSMLOD, smCoords + half2(1, 1) * mSMSize.y).r >= depth)
    shadow[3] = 1.h;

  // lerp between the shadow values to calculate our light amount
  half2 lerps = frac(smCoords * mSMSize.x);
  return lerp(lerp(shadow[0], shadow[1], lerps.x), lerp(shadow[2], shadow[3], lerps.x), lerps.y);
}

/*half CalcShadowVariance(in half2 smCoords, const in float fZ, float fOffset)
{
  float fShadow[9];  
  fShadow[0] = 0.0625f * ((tex2D(samplerSMLOD, smCoords.xy + float2(-fOffset, -fOffset) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[1] = 0.125f * ((tex2D(samplerSMLOD, smCoords.xy + float2(0,        -fOffset) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[2] = 0.0625f * ((tex2D(samplerSMLOD, smCoords.xy + float2(fOffset,  -fOffset) ) < fZ) ? 0.0f : 1.0f);  

  fShadow[3] = 0.125f * ((tex2D(samplerSMLOD, smCoords.xy + float2(-fOffset, 0) ) < fZ) ? 0.0f : 1.0f);
  fShadow[4] = 0.25f * ((tex2D(samplerSMLOD, smCoords.xy + float2(0,        0) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[5] = 0.125f * ((tex2D(samplerSMLOD, smCoords.xy + float2(fOffset,  0) ) < fZ) ? 0.0f : 1.0f);  

  fShadow[6] = 0.0625f * ((tex2D(samplerSMLOD, smCoords.xy + float2(-fOffset, fOffset) ) < fZ) ? 0.0f : 1.0f);
  fShadow[7] = 0.125f * ((tex2D(samplerSMLOD, smCoords.xy + float2(0,        fOffset) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[8] = 0.0625f * ((tex2D(samplerSMLOD, smCoords.xy + float2(fOffset,  fOffset) ) < fZ) ? 0.0f : 1.0f);  

  half fShadowAmount = fShadow[0]+fShadow[1]+fShadow[2]+fShadow[3]+fShadow[4]+fShadow[5]+fShadow[6]+fShadow[7]+fShadow[8];
  return fShadowAmount;
}*/

half CalcShadowVariance(in half2 smCoords, const in float fZ)
{
  float fDepthShift = 0.5;

  // Variance shadow fetching
  float4 moments = tex2D(samplerSMLOD, smCoords);

  // Variance shadow mapping
  float M = moments.r; //mean

  float E_x2 = moments.g;
  float Ex_2 = M * M;
  float variance = (E_x2 - Ex_2); //decrease range of variance to increase precision for 16-bit formats

  //M += fDepthShift; 

  float m_d = M - fZ;
  float p_max = variance / (variance + m_d * m_d);

  // Standard shadow map comparison
  float lit_factor = m_d > 0;

  //select properly shadow region because of one-tailed version of inequality
  half fShadowAmount = max(lit_factor, p_max);
  return fShadowAmount;
}

half4 _Main_PS(const in vertexOutput IN, const int nNumSamples)
{
	// calc world pos
  float cameraZ = tex2D(samplerSceneZ, IN.Tex0_ShadowBias.xy).r;
#ifdef OPENGL
  cameraZ = mSMSize.w/(cameraZ*2.0f+mSMSize.z);
#endif

  half3 rayVector = IN.CameraRay * cameraZ;
  half3 worldPos = (half3)vCEye.xyz + rayVector;
  
  float4 shadowTC = MUL_VEC_MAT(float4(worldPos, 1.f), mShadowWLP0);

  half2 smCoords = shadowTC.xy / shadowTC.w;
  //smCoords.y = 1 - smCoords.y;

  float lightDepth = shadowTC.z / shadowTC.w * 0.5f + 0.5f - IN.Tex0_ShadowBias.z;

  half shadow;

#ifdef SHADOW_VARIANCE
  shadow = CalcShadowVariance(smCoords, lightDepth);
#else	
 #if RT_PROFILE > QUALITY_MEDIUM
  if(nNumSamples != 0)
  {
    half2 vNoisePos = IN.noisePosVec.xy + IN.noisePosVec.zw * cameraZ;
    shadow = CalcShadowPoissonFilter(smCoords, lightDepth, vNoisePos, IN.kernelSize, nNumSamples);
  }
  else
    shadow = CalcShadowBilerp(smCoords, lightDepth);
 #else
  shadow = CalcShadowBilerp(smCoords, lightDepth);
 #endif
#endif


#ifdef RT_POINT_SHADOWMAP
  const float minRange = 0.2f;
  half shadowDarkness = length(smCoords - half2(0.5, 0.5));
  shadowDarkness = smoothstep(minRange, 0.25f, shadowDarkness - minRange);
  shadow = shadow + shadowDarkness;//  + shadowDarkness;
#else
  // fading
  #define FADE_RANGE .2h
  half shadowDarkness = (length(rayVector) - (half)vShadowParams.x * (1.h - FADE_RANGE)) / ((half)vShadowParams.x * FADE_RANGE);
  shadow = max(shadowDarkness, shadow);
#endif
  return shadow;
}

half4 Main_PSJit(in vertexOutput IN) : COLOR0
{
  return _Main_PS(IN, MAX_SHADOW_SAMPLES_NUM);
}
half4 Main_PSHJit(in vertexOutput IN) : COLOR0
{
  return _Main_PS(IN, MAX_SHADOW_SAMPLES_NUM/2);
}
half4 Main_PS0(in vertexOutput IN) : COLOR0
{
  return _Main_PS(IN, 0);
}

half4 Dummy_PS(in vertexOutput IN) : COLOR0
{
  return half4(0, 0, 0, 0);
}

struct VS_INPUT_TC
{
  float4 Pos          : POSITION;
  float2 tc0          : TEXCOORD0;	
};
struct VS_OUTPUT_TC
{
  float4 HPosition   : POSITION;
	float2 tc0   : TEXCOORD0;	
};

VS_OUTPUT_TC Quad_VS(in VS_INPUT_TC In)
{
  VS_OUTPUT_TC Out;
  Out.HPosition = In.Pos;
  Out.tc0 = In.tc0;
  return Out;
}

half4 Blur_PS(in VS_OUTPUT_TC In) : COLOR0
{
  float2 Offset = F_ScreenSize.xy * 2;
  
  float4 sum = tex2D(sampler0, In.tc0);
  sum += tex2D(sampler0, In.tc0 + Offset.xy);
  sum += tex2D(sampler0, In.tc0 + float2( Offset.x, -Offset.y));
  sum += tex2D(sampler0, In.tc0 + float2(-Offset.x,  Offset.y));
  sum += tex2D(sampler0, In.tc0 - Offset.xy);
  return sum / 5;
}

float4 Combine_PS(in VS_OUTPUT_TC In) : COLOR0
{
  half4 TerrainMap = tex2D(sampler0, In.tc0);
  half4 GeneralMap = tex2D(sampler1, In.tc0);
  
  half4 res = min(TerrainMap, GeneralMap);
  
  return res;
}

technique MergeHQ
{
  pass StencilPrepass
  {
    VertexShader = compile VS_SHADER_TYPE StencilPrepass_VS();
    PixelShader  = compile PS_SHADER_TYPE Dummy_PS();
  }
  pass MainPass 
  {
    VertexShader = compile VS_SHADER_TYPE ScreenSpace_VSHJit();
    PixelShader  = compile PS_SHADER_TYPE Main_PSHJit();
  }
}

technique DefaultMerge
{
  pass StencilPrepass
  {
    VertexShader = compile VS_SHADER_TYPE StencilPrepass_VS();
    PixelShader  = compile PS_SHADER_TYPE Dummy_PS();
  }
  pass MainPass
  {
    VertexShader = compile VS_SHADER_TYPE ScreenSpace_VS0();
    PixelShader  = compile PS_SHADER_TYPE Main_PS0();
  }
}
technique MergeHQSM3
{
  pass StencilPrepass
  {
    VertexShader = compile VS_SHADER_TYPE StencilPrepass_VS();
    PixelShader  = compile PS_SHADER_TYPE Dummy_PS();
  }
  pass MainPass 
  {
    VertexShader = compile VS_SHADER_TYPE ScreenSpace_VSJit();
    PixelShader  = compile PS_SHADER_TYPE Main_PSJit();
  }
}

technique Blur
{
  pass MainPass 
  {
    VertexShader = compile VS_SHADER_TYPE Quad_VS();
    PixelShader  = compile PS_SHADER_TYPE Blur_PS();
  }
}

technique Combine
{
  pass MainPass 
  {
    VertexShader = compile VS_SHADER_TYPE Quad_VS();
    PixelShader  = compile PS_SHADER_TYPE Combine_PS();
  }
}
