#include "constants.fxh"
#include "lighting.fxh"

float4 g_vCameraAngles;

#define MAX_NUM_LEAF_ANGLES 10
float4x4      g_avLeafAngles[MAX_NUM_LEAF_ANGLES]; // each element: .x = rock angle, .y = rustle angle
                                                 // each element is a float4, even though only a float2 is needed, to facilitate
                                                 // fast uploads on all platforms (one call to upload whole array)
float4      g_vLeafAngleScalars = {0,0,0,0};     // each tree model has unique scalar values: .x = rock scalar, .y = rustle scalar
float4x4    g_mLeafUnitSquare =                  // unit leaf card that's turned towards the camera and wind-rocked/rustled by the
            {                                    // vertex shader.  card is aligned on YZ plane and centered at (0.0f, 0.0f, 0.0f)
                float4(0.0f, 0.5f, 0.5f, 0.0f), 
                float4(0.0f, -0.5f, 0.5f, 0.0f), 
                float4(0.0f, -0.5f, -0.5f, 0.0f), 
                float4(0.0f, 0.5f, -0.5f, 0.0f)
            };


#define NUM_WIND_MATRICES 10
float4x3	WindMatrices[NUM_WIND_MATRICES];

float3 WindEffect(float3 vPosition, float2 vWindInfo)
{
    float2 vWeights = frac(vWindInfo.xy);

    float2 vIndices = (vWindInfo - vWeights) * 0.05f * NUM_WIND_MATRICES;
    float3 vWindEffect = lerp(vPosition.xyz, mul(vPosition, WindMatrices[int(vIndices.x)]), vWeights.x);
	return lerp(vWindEffect, mul(vWindEffect, WindMatrices[int(vIndices.y)]), vWeights.y);
}

float3 LightDiffuse(float3 vVertex,
                    float3 vNormal,
                    float3 vLightDir,
                    float3 vLightColor,
                    float3 vDiffuseMaterial)
{
    return vDiffuseMaterial * vLightColor * max(dot(vNormal, vLightDir), 0.0f);
}

float3x3 RotationMatrix_zAxis(float fAngle)
{
    // compute sin/cos of fAngle
    float2 vSinCos;
    sincos(fAngle, vSinCos.x, vSinCos.y);
    
    return float3x3(vSinCos.y, -vSinCos.x, 0.0f, 
                    vSinCos.x, vSinCos.y, 0.0f, 
                    0.0f, 0.0f, 1.0f);
}

float3x3 RotationMatrix_yAxis(float fAngle)
{
    // compute sin/cos of fAngle
    float2 vSinCos;
    sincos(fAngle, vSinCos.x, vSinCos.y);
    
    return float3x3(vSinCos.y, 0.0f, vSinCos.x,
                    0.0f, 1.0f, 0.0f,
                    -vSinCos.x, 0.0f, vSinCos.y);
}

float3x3 RotationMatrix_xAxis(float fAngle)
{
    // compute sin/cos of fAngle
    float2 vSinCos;
    sincos(fAngle, vSinCos.x, vSinCos.y);
    
    return float3x3(1.0f, 0.0f, 0.0f,
                    0.0f, vSinCos.y, -vSinCos.x,
                    0.0f, vSinCos.x, vSinCos.y);
}



void vs_Main(	in float4   vPosition  : POSITION,  // xyz = position, w = corner index
                in float4   vTexCoord0 : TEXCOORD0, // xy = diffuse texcoords, zw = compressed wind parameters
                in float4   vTexCoord1 : TEXCOORD1, // .x = width, .y = height, .z = pivot x, .w = pivot.y
                in float4   vTexCoord2 : TEXCOORD2,
                in float3   vNormal    : NORMAL,
                out float4  vOutPos    : POSITION,
                out float4  vOutColor  : COLOR,
                out float2  vOutUv0    : TEXCOORD0 )
{
    // define attribute aliases for readability
    float fAzimuth = g_vCameraAngles.x;      // camera azimuth for billboarding
    float fPitch = g_vCameraAngles.y;        // camera pitch for billboarding
    float2 vSize = vTexCoord1.xy;            // leaf card width & height
    int nCorner = vPosition.w;               // which card corner this vertex represents [0,3]
    float fRotAngleX = vTexCoord2.x;         // angle offset for leaf rocking (helps make it distinct)
    float fRotAngleY = vTexCoord2.y;         // angle offset for leaf rustling (helps make it distinct)
    float fWindAngleIndex = vTexCoord2.z;    // which wind matrix this leaf card will follow
    float2 vPivotPoint = vTexCoord1.zw;      // point about which card will rock and rustle
    float fDimming = vTexCoord2.w;           // interior leaves are darker (range = [0.0,1.0])
    float2 vWindParams = vTexCoord0.zw;      // compressed wind parameters

    //vPosition.xz = float2(dot(g_vTreeRotationTrig.ywz, vPosition.xyz), dot(g_vTreeRotationTrig.xwy, vPosition.xyz));
    //vNormal.xz = float2(dot(g_vTreeRotationTrig.ywz, vNormal.xyz), dot(g_vTreeRotationTrig.xwy, vNormal.xyz));
    
    vPosition.xyz = mul(vPosition.xyz, (float3x3)PI_mW);
    vPosition.xyz = WindEffect(vPosition.xyz, vWindParams);
    vPosition.xyz += PI_mW._41_42_43;
    
    float2 vLeafRockAndRustle = g_vLeafAngleScalars.xy * g_avLeafAngles[fWindAngleIndex]._11_12;
    //float2 vLeafRockAndRustle = {0,0};
    float3 vCorner = g_mLeafUnitSquare[nCorner].yzx * vSize.xyx;

       
    float3x3 matRotation = RotationMatrix_yAxis(fAzimuth + fRotAngleX);
    matRotation = mul(matRotation, RotationMatrix_zAxis(fPitch + fRotAngleY));
    matRotation = mul(matRotation, RotationMatrix_yAxis(-vLeafRockAndRustle.y));
    matRotation = mul(matRotation, RotationMatrix_xAxis(vLeafRockAndRustle.x));
    vCorner = mul(matRotation, vCorner);
    
    vPosition = mul(float4(vPosition.xyz,1), C_V);
    vPosition.xyz += vCorner;

    // translate tree into position (must be done after the rotation)
    //vPosition.xyz += g_vTreePos.xyz;

    float f0 = 0;

    // compute the leaf lighting (not using normal mapping, but per-vertex lighting)
    //vOutColor.rgb = ArrayTest[f0]._11_22_33;//fDimming * LightDiffuse(vPosition.xyz, vNormal.xyz, LGetDir(0).xyz, LGetDiffuse(0).rgb, vWorldAmbient.rgb);
    vOutColor.rgb = F_WorldAmbient.xyz;
#if defined(RT_NUM_LIGHTS)
		vOutColor.rgb += saturate(dot(vNormal,LGetDir(0).xyz))*LGetDiffuse(0).xyz;
#endif
    vOutColor.a = 1.0f;

    // project position to the screen
    vOutPos = mul(vPosition, mP);
    
    // pass through other texcoords exactly as they were received
    vOutUv0.xy = vTexCoord0.xy;
}


//-----------------------------------
technique Default
{
    pass p0 
    {		
		VertexShader = compile VS_SHADER_TYPE vs_Main();
		
		Texture[0] = <Texture0>;
		// Just use the color


		ColorArg1[0] = Diffuse;
		ColorArg2[0] = Texture;
		AlphaArg1[0] = Diffuse;
		AlphaArg2[0] = Texture;
		ColorOp[0] = Modulate;
		AlphaOp[0] = Modulate;
		ColorOp[1] = Disable;
		AlphaOp[1] = Disable;		
    }
}