#define fragInput vertexOutput
#define RT_TEX_SCROLL
#define FRAG_CUSTOM_END
#define USE_SCREENCOORDS

#include "transform.fxh"
#include "fragDef.fxh"

void frag_custom_per_light(inout fragPass pPass, inout fragLightPass pLight)
{
}

void frag_custom_end(inout fragPass pPass, inout FIXED3 cFinal)
{
  half3 cDiffuse = pPass.cDiffuseMap.xyz * pPass.cVertexColor.rgb;
	cDiffuse *= materialColor.rgb;

	// refraction
#if RT_PROFILE > QUALITY_MEDIUM
	half2 vTexCoord = pPass.IN.ScreenCoordProj.xy / pPass.IN.ScreenCoordProj.z;

	#if defined(BUMP_DIFFUSE)
		half2 vDisplacement = tex2D(SpecularSampler, pPass.IN.Tex0).rg * 2.h - 1.h;
	#else
		half2 vDisplacement = (half2(sin(pPass.IN.WorldPos.y * 25.f + F_Time.x * 2.5f), cos(pPass.IN.WorldPos.z * 28.f - F_Time.x * 1.7f) * .9f)) * .01f;
	#endif

	// depth-based effects
	float fDepth = tex2D(DepthBufferSampler, vTexCoord);
	float fWorldCameraDepth = fDepth;
	float fDistance = (fWorldCameraDepth - pPass.IN.WorldPos.w) * C_FarClip.x;
	// underwater fog
	half fFog = saturate(fDistance / 1.0f);
	// Screen-space displacement amplitude compensation
	vDisplacement *= saturate(2.f / (pPass.IN.WorldPos.w * C_FarClip.x));
	// Depth-based ray deviation
	vDisplacement *= saturate(fDistance);

	vTexCoord += vDisplacement;

	half3 vRefraction = tex2D(BackBufferSampler, vTexCoord);

	cDiffuse = lerp(vRefraction, cDiffuse, fFog);
#endif

  cFinal.rgb += cDiffuse;

	// reflection
#if RT_PROFILE > QUALITY_LOW
#if defined(RT_SPECULAR) && defined(RT_ENVMAP)
	// do a reflection cubemap lookup
	half3 vReflections = texCUBE(EnvMapSampler, pPass.vReflVec).rgb;

	// affect specular by a Fresnel term
	if( pPass.nQuality >= QUALITY_MEDIUM )
	{
		half fFresnelTerm = GetFresnel(pPass.fNdotE, pPass.nQuality, false);
		vReflections *= fFresnelTerm * Ks;
	}

	cFinal.rgb += vReflections;
#endif 
#endif


#if defined(RT_FOG) && RT_PROFILE > QUALITY_VERYLOW && defined(SM_30)	// disable fog for blending
	pPass.IN.FogColor.xyz = 0;
	cFinal.rgb *= pPass.IN.FogColor.a;
	pPass.IN.FogColor.a = 1.h;
#endif
}

#include "fragLib.fxh"

half4 PS_StillWater(const in fragInput IN) : COLOR
{
  // Initialize fragPass structure
  fragPass pPass;
#ifndef OPENGL
  pPass = (fragPass)0;
#endif
  frag_unify(pPass, IN);  
  half4 cFinal = frag_shared_output(pPass);
  //cFinal.w = fOpacity;

  return cFinal;
}


//========================================================================================================================================================================

struct v2fReflect
{
  float4 Position  : POSITION;  // in projection space
  float4 Wave0     : TEXCOORD0;  //xy: wave0, wz: wave1
  float4 Wave1     : TEXCOORD1;  //xy: wave2, wz: wave3

  float3 projTC    : TEXCOORD2;
  float4 viewVec   : TEXCOORD3;  
  float3 cameraDir : TEXCOORD4;  

  float4 FogColor  : TEXCOORD5;
#if defined(TEST) && RT_PROFILE > QUALITY_HIGH
  float3 reflectTC : TEXCOORD6;
#endif
};

extern const float WavesTiling;
extern const float WavesSpeed;
extern const float WavesAmount;
extern const float WavesSize;
extern const float BumpScale;
extern const float ReflectionAmount;
extern const float SoftIntersectionFactor;
extern const float FresnelBias;
extern const float RefrBumpScale;
extern const float WindDirection;
extern const float4 WaterColor;

half GetFresnelOpt(half NdotI, half bias, half power)
{
  half facing = (1.0 - NdotI);
  return saturate(bias + (1-bias)*pow(facing, power));
}

///////////////// vertex shader //////////////////
v2fReflect StillWaterVS(in VS_IN In)
{
  v2fReflect OUT;
  int nQuality = GetShaderQuality(); 

  OUT = (v2fReflect)0; 
  
  float4 OceanGenParams = float4(F_Time.x*0.0125, F_Time.x*0.025, F_Time.x*0.05, F_Time.x);

  // Position in screen space.
  float3 WorldPos = WorldPosition(In);
  OUT.Position = mul(float4(WorldPos, 1), C_VP);
    
  // Output bump layers texture coordinates
  float2 vTranslation= float2(OceanGenParams.y * WavesSpeed, 0);                       
  
  float2 vTranslation0 = vTranslation;      
  float2 vTranslation1 = vTranslation;
  
  float2 vTex = float4(In.Tex0.xy, 0, 1)*4.0;
  vTex *= WavesTiling;

  // Scale texture coordinates to get mix of low/high frequency details      
  float fLowSpecScale = (64.0 / 256.0);

#if !NO_FLOW  
  if (nQuality > QUALITY_LOW)
  {
    OUT.Wave0.xy = vTex.xy*4 + vTranslation0.xy*2;
    OUT.Wave0.wz = vTex.xy*2.0 + vTranslation1.xy*3;
    OUT.Wave1.xy = (vTex.xy*2.0*4 + vTranslation0.xy*2.0);
    OUT.Wave1.wz = (vTex.xy*4.0*4 + vTranslation1.xy*3.0);
  }
  else
  {
    OUT.Wave0.xy = vTex.xy*fLowSpecScale + OceanGenParams.x*0.5;
    OUT.Wave0.wz = vTex.xy*0.9*fLowSpecScale  - OceanGenParams.x*0.5;
    OUT.Wave1.xy = (vTex.xy*2.0*fLowSpecScale + OceanGenParams.x);
    OUT.Wave1.wz = (vTex.xy*2.0*fLowSpecScale- OceanGenParams.x);
  }
#else
  if (nQuality > QUALITY_LOW)
  {
    OUT.Wave0.xy = vTex.xy*2 * 4;
    OUT.Wave0.wz = vTex.xy*3 * 4;
    OUT.Wave1.xy = (vTex.xy*4*4 + vTranslation0.xy*2.0*4);
    OUT.Wave1.wz = (vTex.xy*5 *4- vTranslation0.xy*2.0*4); ; 
  }
  else
  {
    OUT.Wave0.xy = vTex.xy*2*fLowSpecScale  + OceanGenParams.x;
    OUT.Wave0.wz = vTex.xy*1.9*fLowSpecScale  - OceanGenParams.x;
    OUT.Wave1.xy = (vTex.xy*4*fLowSpecScale + OceanGenParams.x*2);
    OUT.Wave1.wz = (vTex.xy*5*fLowSpecScale - OceanGenParams.x*2); ;;
  }
#endif  
  
  float3 vNorm = float3(0,0,1);  
  
  // Output eye/light vector
  float3 viewVec = WorldPos.xyz - C_Eye.xyz;
  OUT.viewVec.xyz = viewVec.xyz;
  OUT.viewVec.w = OUT.Position.z * C_FarClip.y;
  OUT.cameraDir.xyz = C_VI[2].xyz;

#if defined(TEST) && RT_PROFILE > QUALITY_HIGH
  {
    float4 vReflPos = float4(WorldPos.xyz, 1);
    OUT.reflectTC.xyz = mul(vReflPos.xyzw, C_VolReflectMatrix).xyw;
  }
#endif  

  // Output projected coordinates
  OUT.projTC = HPosToScreenTC(OUT.Position);

#if defined (RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  OUT.FogColor = Fog(length(C_VI[3].xyz - WorldPos));
#endif

  return OUT;
}

// Get vertex from screen space into world spasace
void GetGridVertexPos(inout float4 vPos, inout float fDistToCam )
{
  fDistToCam = 0;
}

// Get waves
void GetWave(inout float3 vPos, in float2 FlowDir, in VS_IN IN, bool bOneSample = true)
{
  if (GetShaderQuality() <= QUALITY_LOW)
    return;

  // constant to scale down values a bit
  const float fAnimAmplitudeScale = 1.0 / 5.0;

  float3 vPosOrig = vPos;
  float fPhaseTest = length(vPosOrig.xz);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Water ocean animation

  // Each wave as ~doubled frequency/phase
  float4 vFrequencies = 5*WavesSpeed * float4(0.233, 0.455, 0.6135, -0.1467);
  float4 vPhases = WavesAmount * float4(0.1, 0.159, 0.557, 0.2199);
  float4 vAmplitudes = WavesSize * float4(1, 0.5, 0.25, 0.5); 

  float4 vWaveFreq = F_Time.x * vFrequencies;                                   // 1 inst
  
  float4 vCosPhase = (fPhaseTest + vPosOrig.x)* vPhases;
  float4 vCosWave = cos(vWaveFreq * FlowDir.x + vCosPhase);
  
  float4 vSinPhase = (fPhaseTest + vPosOrig.z)* vPhases; 
  float4 vSinWave = sin(vWaveFreq * FlowDir.y + vSinPhase);

  float fAnimSum = dot(vCosWave, vAmplitudes) + dot(vSinWave, vAmplitudes);                 // 3 inst

  // temporary solution, quality levels not working atm on xbox
	if (GetShaderQuality() > QUALITY_LOW) 
  	vPos.y += fAnimSum * fAnimAmplitudeScale;                                                 // 1 inst

  //float3 v0 = tex2Dlod_bilinear(fftTexSampler, (vFakePos.xz)*0.0125*OceanParams0.w * 1.25, 64, 1.0/64); //*3
  //vPos.xyz += v0.xyz * 0.06 * OceanParams1.x * float3(1.5,1.5,1); 
}

v2fReflect _OceanVS(in VS_IN In, bool bSimple)
{
  v2fReflect OUT;
  int nQuality = GetShaderQuality(); 

  OUT = (v2fReflect)0; 
  
  float4 OceanGenParams = float4(F_Time.x*0.0125, F_Time.x*0.025, F_Time.x*0.05, F_Time.x);

  // Position in screen space.
  float4 WorldPos;
  WorldPos.xyz = WorldPosition(In);
  WorldPos.w = 1;
  
  float fVertexDist = 0;  
  GetGridVertexPos(WorldPos, fVertexDist); // 30 alu
  
  float4 vPosOrig = WorldPos;
  float3 vPos0 = vPosOrig.xyz;

  float2 FlowDir; FlowDir.x = cos(WindDirection); FlowDir.y = sin(WindDirection);
  GetWave(WorldPos.xyz, FlowDir, In, false);
  
  float fDistToCam = length(C_Eye.xyz - WorldPos.xyz);
  // Attenuate wave strength nearby camera  
  float fNearbyCameraAtten = (saturate(fDistToCam*fDistToCam));
  WorldPos.xyz = lerp(vPosOrig.xyz, WorldPos.xyz, saturate(fNearbyCameraAtten));
  
  OUT.Position = mul(WorldPos, C_VP);
    
  // Output bump layers texture coordinates
  float2 vTranslation= float2(OceanGenParams.y * WavesSpeed, 0);                       
  
  float2 vTranslation0 = vTranslation;      
  float2 vTranslation1 = vTranslation;
  
  float2 vTex = float4(In.Tex0.xy, 0, 1)*4.0;
  vTex *= WavesTiling;

  // Scale texture coordinates to get mix of low/high frequency details      
  float fLowSpecScale = (64.0 / 256.0);

#if !NO_FLOW  
  if (nQuality > QUALITY_LOW)
  {
    OUT.Wave0.xy = vTex.xy*4 + vTranslation0.xy*2;
    OUT.Wave0.wz = vTex.xy*2.0 + vTranslation1.xy*3;
    OUT.Wave1.xy = (vTex.xy*2.0*4 + vTranslation0.xy*2.0);
    OUT.Wave1.wz = (vTex.xy*4.0*4 + vTranslation1.xy*3.0);
  }
  else
  {
    OUT.Wave0.xy = vTex.xy*fLowSpecScale + OceanGenParams.x*0.5;
    OUT.Wave0.wz = vTex.xy*0.9*fLowSpecScale  - OceanGenParams.x*0.5;
    OUT.Wave1.xy = (vTex.xy*2.0*fLowSpecScale + OceanGenParams.x);
    OUT.Wave1.wz = (vTex.xy*2.0*fLowSpecScale- OceanGenParams.x);
  }
#else
  if (nQuality > QUALITY_LOW)
  {
    OUT.Wave0.xy = vTex.xy*2 * 4;
    OUT.Wave0.wz = vTex.xy*3 * 4;
    OUT.Wave1.xy = (vTex.xy*4*4 + vTranslation0.xy*2.0*4);
    OUT.Wave1.wz = (vTex.xy*5 *4- vTranslation0.xy*2.0*4); ; 
  }
  else
  {
    OUT.Wave0.xy = vTex.xy*2*fLowSpecScale  + OceanGenParams.x;
    OUT.Wave0.wz = vTex.xy*1.9*fLowSpecScale  - OceanGenParams.x;
    OUT.Wave1.xy = (vTex.xy*4*fLowSpecScale + OceanGenParams.x*2);
    OUT.Wave1.wz = (vTex.xy*5*fLowSpecScale - OceanGenParams.x*2); ;;
  }
#endif  
  
  float3 vNorm = float3(0,0,1);  
  
  // Output eye/light vector
  float3 viewVec = WorldPos.xyz - C_Eye.xyz;
  OUT.viewVec.xyz = viewVec.xyz;
  OUT.viewVec.w = OUT.Position.z * C_FarClip.y;
  OUT.cameraDir.xyz = C_VI[2].xyz;

  // Output projected coordinates
  OUT.projTC = HPosToScreenTC(OUT.Position);

#if defined(TEST) && RT_PROFILE > QUALITY_HIGH
  {
    float4 vReflPos = float4(WorldPos.xyz, 1);
    OUT.reflectTC.xyz = mul(vReflPos.xyzw, C_VolReflectMatrix).xyw;
  }
#endif  

#if defined (RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  OUT.FogColor = Fog(length(C_VI[3].xyz - WorldPos));
#endif

  return OUT;
}

v2fReflect OceanVS(in VS_IN In)
{
  return _OceanVS(In, false);
}
v2fReflect SimpleWaterVS(in VS_IN In)
{
  return _OceanVS(In, true);
}

half4 StillWaterPS(v2fReflect IN) : COLOR
{
  int nQuality = GetShaderQuality(); 
  half3 viewVec = normalize(-IN.viewVec.xyz);  

  float4 OceanGenParams = float4(F_Time.x*0.0125, F_Time.x*0.025, F_Time.x*0.05, F_Time.x);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // generate bump
  //////////////////////////////////////////////////////////////////////////////////////////////////

  half3 bumpNormal = half3(0,0,1);
  half fVariation = 1;

  if (nQuality > QUALITY_LOW)
  {
    // Simple offset mapping - use pom for very hi-spec
    half height = (tex2D(WaterNormalSampler, IN.Wave0.xy*0.33).w*2-1);

    float cam_wave = 0;
    if (nQuality >= QUALITY_HIGH)
    {
      float l = length(IN.viewVec.xyz);
      cam_wave = sin(l*10-OceanGenParams.w*5) * saturate(1 - (l/2.5));
    }

    float2 vDisp = (height+cam_wave) * viewVec.xy * 0.1; //*fVariation;
    fVariation = tex2D(WaterNormalSampler, IN.Wave0.xy*0.05 + vDisp).w;
    fVariation += tex2D(WaterNormalSampler, IN.Wave0.xy*0.1 + fVariation+  vDisp).w;
    
    // swap bump depending on quality settings
    float4 n0 = tex2D(WaterNormalSampler, IN.Wave0.xy*0.33 + vDisp)*2-1;
    float4 n1 = tex2D(WaterNormalSampler, IN.Wave0.xy*0.57 + vDisp)*2-1;
    float4 n2 = tex2D(WaterNormalSampler, IN.Wave0.xy*0.959 + vDisp)*2-1;

    bumpNormal.xyz = n0;
    bumpNormal.xy += n1.xy ;
    bumpNormal.xyz = normalize( bumpNormal.xyz );

    bumpNormal.xy += n2.xy ;
    bumpNormal.xyz = normalize( bumpNormal.xyz  * half3(BumpScale.xx*fVariation, 1));
  }
  else
  {
    half2 bumpColorA = 0;
    half2 bumpColorB = 0;
    half2 bumpColorC = 0;
    half2 bumpColorD = 0;
    half2 bumpLowFreq = 0;
    half2 bumpHighFreq = 0;
    
    // merge big waves
    bumpColorA.xy = tex2D(WaterNormalSampler, IN.Wave0.xy);           
    bumpColorB.xy = tex2D(WaterNormalSampler, IN.Wave0.wz);           
    
    bumpLowFreq.xy = (bumpColorA + bumpColorB);                       // 2 alu  

    // merge small waves
    bumpColorC.xy = tex2D(WaterNormalSampler, IN.Wave1.xy);
    bumpColorD.xy = tex2D(WaterNormalSampler, IN.Wave1.wz);
    
    bumpHighFreq.xy = (bumpColorC.xy + bumpColorD.xy);            // 2 alu

    // merge all waves
    bumpNormal.xy = (bumpLowFreq + bumpHighFreq) * 2.0 - 4.0;                                                     // 1 alu  
    bumpNormal.xy *= BumpScale;                                                                     // 1 alu

    // just assume bump normal z is 1 and renormalize
    normalize(bumpNormal.xyz);
  }

  float3 vn = bumpNormal;
  bumpNormal.y = vn.z;
  bumpNormal.z = vn.y;
  float3 vReflect = (2*dot(viewVec,bumpNormal)*bumpNormal - viewVec);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // water edge softness
  //////////////////////////////////////////////////////////////////////////////////////////////////

  float fA = 1;  
  half2 vProject = IN.projTC.xy/IN.projTC.z;
  float sceneDepth = tex2D(DepthBufferSampler, vProject).r * C_FarClip.x;
  float waterDepth = IN.projTC.z;
  float fDistance = sceneDepth - waterDepth;
  float depth = max(fDistance, 0) * C_FarClip.y;
  half softIntersect = saturate(SoftIntersectionFactor * depth);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Get reflection
  //////////////////////////////////////////////////////////////////////////////////////////////////

  half3 reflectColor = 1.0;
#if defined(TEST) && RT_PROFILE > QUALITY_HIGH
  {
    float2 vReflect = IN.reflectTC.xy / IN.reflectTC.z + vn.xy;
    reflectColor = tex2D(WaterReflectionSampler, vReflect.xy);
  }
#else  
  half4 cCubeColor = texCUBE(EnvMapSampler, vReflect.xyz);
  reflectColor = cCubeColor;
#endif  
  
  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Compute coeficients      
  //////////////////////////////////////////////////////////////////////////////////////////////////
            
  half NdotI = (dot(viewVec, bumpNormal.xyz));   // taking absolute value, so that refraction looks correct inside water          
  half fInsideReflection = (1 - saturate(1000.0 * (NdotI+0.1)));
  NdotI = abs(NdotI);   // taking absolute value, so that refraction looks correct inside water          
  half fFresnelTerm = ReflectionAmount * GetFresnelOpt(NdotI, FresnelBias, 5);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Get refraction
  //////////////////////////////////////////////////////////////////////////////////////////////////

  half2 refrNewst = vProject; // reduce refraction near beaches

  float2 refrTC = bumpNormal.xz;
  if (nQuality >= QUALITY_HIGH)
  {
    refrTC *= (RefrBumpScale + 0.25*fInsideReflection)*softIntersect; //*(saturate(0.0005/IN.viewVec.z)); // Difraction amount always 0.1
    // ultra-high quality
    float3 depthRefr = tex2D(DepthBufferSampler, refrNewst + refrTC.xy * 1.05).r;
    depthRefr.y = tex2D(DepthBufferSampler, refrNewst + refrTC.xy).r;
    depthRefr.z = tex2D(DepthBufferSampler, refrNewst + refrTC.xy * 0.95).r;

    half3 fRefractionMask =  (IN.viewVec.www < depthRefr.xyz);

    // Apply refraction mask to bump offsets
    refrTC *= fRefractionMask.x * fRefractionMask.y * fRefractionMask.z; 	  
  }
  else
  {
    refrTC *= RefrBumpScale * (saturate(0.0005/IN.viewVec.w)); // Difraction amount always 0.1

    float3 depthRefr = tex2D(DepthBufferSampler, refrNewst + refrTC.xy * 1.15).r;
    half fRefractionMask = IN.viewVec.w <depthRefr.x;
    // Apply refraction mask to bump offsets
    refrTC *= saturate(fRefractionMask);
  }

  half3 refractColor = 0;
  if (nQuality >= QUALITY_HIGH)
  {
    refractColor.x = tex2D(BackBufferSampler, refrNewst + refrTC * 1.05).x;
    refractColor.y = tex2D(BackBufferSampler, refrNewst + refrTC).y;
    refractColor.z = tex2D(BackBufferSampler, refrNewst + refrTC * 0.95).z;
  } 
  else
  {
    // no color dispersion for lower specs
    refractColor.xyz = tex2D(BackBufferSampler, refrNewst + refrTC * 1.15 ).xyz;
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Compute sun specular
  //////////////////////////////////////////////////////////////////////////////////////////////////

#if RT_SPECULAR
  if (nQuality >= QUALITY_MEDIUM)
  {
    half fFoamGloss = tex2D(SpecularSampler, IN.Wave0.wz*0.1 + bumpNormal.xy).x;                // 1 alu
    fFoamGloss += tex2D(SpecularSampler, IN.Wave0.xy*0.2 + bumpNormal.xy).x;                     // 2 alu, 1 mov
    //fFoamGloss *= 0.5;
    fFoamGloss = saturate(fFoamGloss*0.5+0.5);

    // Sun glow/shine
    half RdoTL = saturate(dot(vReflect.xyz, L_LightDir0)); //*selfshadowK;

    half sunSpecular = pow(RdoTL , 512);                        
    half3 vSunGlow = saturate(L_LightDir0.z*4) * sunSpecular * L_LightDiffuse0.xyz * SunMultiplier * fFoamGloss;
    //vSunGlow += vSunGlow *25 *saturate(fFoamGloss-0.55)* L_LightDiffuse0.xyz;

    // Add sun specular term
    reflectColor.xyz += vSunGlow.xyz;                    
  }
#endif

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Underwater fog
  //////////////////////////////////////////////////////////////////////////////////////////////////
  if (nQuality > QUALITY_MEDIUM)
  {
    half3 cFogColor = materialColor.rgb * float3(0.45, 0.45, 0.5);
    // underwater fog
    half fFog = saturate(fDistance / 10.0f);
    fFog = 0;
    refractColor = lerp(refractColor, cFogColor, fFog);
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Final composition
  //////////////////////////////////////////////////////////////////////////////////////////////////

  // Compute final water color             
  half3 final = lerp(refractColor.xyz,  reflectColor, saturate(fFresnelTerm)) ;
  //half3 final = refractColor.xyz +  reflectColor * fFresnel;

  // soft intersection with camera
  if (nQuality > QUALITY_LOW)
    fA *= saturate((IN.projTC.z - C_FarClip.z) * 5);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Blend fog
  //////////////////////////////////////////////////////////////////////////////////////////////////
  frag_fog_setup(IN.FogColor, float4(final, fA));

  //final.xyz = reflectColor.xyz;
  //fA = 1;
  return half4(final.xyz, fA);
}

half4 SimpleWaterPS(v2fReflect IN) : COLOR
{
  int nQuality = GetShaderQuality(); 
  half3 viewVec = normalize(-IN.viewVec.xyz);  

  float4 OceanGenParams = float4(F_Time.x*0.0125, F_Time.x*0.025, F_Time.x*0.05, F_Time.x);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // generate bump
  //////////////////////////////////////////////////////////////////////////////////////////////////

  half3 bumpNormal = half3(0,0,1);
  half fVariation = 1;

  if (nQuality > QUALITY_LOW)
  {
    // Simple offset mapping - use pom for very hi-spec
    half height = (tex2D(WaterNormalSampler, IN.Wave0.xy*0.33).w*2-1);

    float cam_wave = 0;
    if (nQuality >= QUALITY_HIGH)
    {
      float l = length(IN.viewVec.xyz);
      cam_wave = sin(l*10-OceanGenParams.w*5) * saturate(1 - (l/2.5));
    }

    float2 vDisp = (height+cam_wave) * viewVec.xy * 0.1; //*fVariation;
    fVariation = tex2D(WaterNormalSampler, IN.Wave0.xy*0.05 + vDisp).w;
    fVariation += tex2D(WaterNormalSampler, IN.Wave0.xy*0.1 + fVariation+  vDisp).w;
    
    // swap bump depending on quality settings
    float4 n0 = tex2D(WaterNormalSampler, IN.Wave0.xy*0.33 + vDisp)*2-1;
    float4 n1 = tex2D(WaterNormalSampler, IN.Wave0.xy*0.57 + vDisp)*2-1;
    float4 n2 = tex2D(WaterNormalSampler, IN.Wave0.xy*0.959 + vDisp)*2-1;

    bumpNormal.xyz = n0;
    bumpNormal.xy += n1.xy ;
    bumpNormal.xyz = normalize( bumpNormal.xyz );

    bumpNormal.xy += n2.xy ;
    bumpNormal.xyz = normalize( bumpNormal.xyz  * half3(BumpScale.xx*fVariation, 1));
  }
  else
  {
    half2 bumpColorA = 0;
    half2 bumpColorB = 0;
    half2 bumpColorC = 0;
    half2 bumpColorD = 0;
    half2 bumpLowFreq = 0;
    half2 bumpHighFreq = 0;
    
    // merge big waves
    bumpColorA.xy = tex2D(WaterNormalSampler, IN.Wave0.xy);           
    bumpColorB.xy = tex2D(WaterNormalSampler, IN.Wave0.wz);           
    
    bumpLowFreq.xy = (bumpColorA + bumpColorB);                       // 2 alu  

    // merge small waves
    bumpColorC.xy = tex2D(WaterNormalSampler, IN.Wave1.xy);
    bumpColorD.xy = tex2D(WaterNormalSampler, IN.Wave1.wz);
    
    bumpHighFreq.xy = (bumpColorC.xy + bumpColorD.xy);            // 2 alu

    // merge all waves
    bumpNormal.xy = (bumpLowFreq + bumpHighFreq) * 2.0 - 4.0;                                                     // 1 alu  
    bumpNormal.xy *= BumpScale;                                                                     // 1 alu

    // just assume bump normal z is 1 and renormalize
    normalize(bumpNormal.xyz);
  }

  float3 vn = bumpNormal;
  bumpNormal.y = vn.z;
  bumpNormal.z = vn.y;
  float3 vReflect = (2*dot(viewVec,bumpNormal)*bumpNormal - viewVec);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // water edge softness
  //////////////////////////////////////////////////////////////////////////////////////////////////

  float fA = 1;  
  half2 vProject = IN.projTC.xy/IN.projTC.z;
  float sceneDepth = tex2D(DepthBufferSampler, vProject).r * C_FarClip.x;
  float waterDepth = IN.projTC.z;
  float fDistance = sceneDepth - waterDepth;
  float depth = max(fDistance, 0) * C_FarClip.y;
  half softIntersect = saturate(SoftIntersectionFactor * depth);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Get reflection
  //////////////////////////////////////////////////////////////////////////////////////////////////

  half3 reflectColor = 1.0;
#if defined(TEST) && RT_PROFILE > QUALITY_HIGH
  {
    float2 vReflect = IN.reflectTC.xy / IN.reflectTC.z + bumpNormal.xy*0.025f;
    reflectColor = tex2D(WaterReflectionSampler, vReflect.xy);
  }
#else  
  half4 cCubeColor = texCUBE(EnvMapSampler, vReflect.xyz);
  reflectColor = cCubeColor;
#endif  
  
  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Compute coeficients      
  //////////////////////////////////////////////////////////////////////////////////////////////////
            
  half NdotI = (dot(viewVec, bumpNormal.xyz));   // taking absolute value, so that refraction looks correct inside water          
  half fInsideReflection = (1 - saturate(1000.0 * (NdotI+0.1)));
  NdotI = abs(NdotI);   // taking absolute value, so that refraction looks correct inside water          
  half fFresnelTerm = GetFresnelOpt(NdotI, FresnelBias, 5) * (ReflectionAmount + 1);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Compute sun specular
  //////////////////////////////////////////////////////////////////////////////////////////////////

#if RT_SPECULAR
  if (nQuality >= QUALITY_MEDIUM)
  {
    half fFoamGloss = tex2D(SpecularSampler, IN.Wave0.wz*0.1 + bumpNormal.xy).x;                // 1 alu
    fFoamGloss += tex2D(SpecularSampler, IN.Wave0.xy*0.2 + bumpNormal.xy).x;                     // 2 alu, 1 mov
    //fFoamGloss *= 0.5;
    fFoamGloss = saturate(fFoamGloss*0.5+0.5);

    // Sun glow/shine
    half RdoTL = saturate(dot(vReflect.xyz, L_LightDir0)); //*selfshadowK;

    half sunSpecular = pow(RdoTL , 512);                        
    half3 vSunGlow = saturate(L_LightDir0.z*4) * sunSpecular * L_LightDiffuse0.xyz * SunMultiplier * fFoamGloss;
    //vSunGlow += vSunGlow *25 *saturate(fFoamGloss-0.55)* L_LightDiffuse0.xyz;

    // Add sun specular term
    reflectColor.xyz += vSunGlow.xyz;                    
  }
#endif

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Final composition
  //////////////////////////////////////////////////////////////////////////////////////////////////

  // Compute final water color             
  half3 final = lerp(WaterColor.xyz, reflectColor, saturate(fFresnelTerm)) ;

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Blend fog
  //////////////////////////////////////////////////////////////////////////////////////////////////
  frag_fog_setup(IN.FogColor, float4(final, fA));

  //final.xyz = reflectColor.xyz;
  //fA = 1;
  return half4(final.xyz, 1);
}

//=============================================================================================================================

technique StillWater < string name = "Material still water."; >
{
  pass p0
  {
    //VertexShader = compile VS_SHADER_TYPE UberVS();
    //PixelShader = compile PS_SHADER_TYPE PS_StillWater();
    VertexShader = compile VS_SHADER_TYPE StillWaterVS();
    PixelShader = compile PS_SHADER_TYPE StillWaterPS();
    DEFAULT_STATES
  }
}

technique SimpleWater < string name = "Material simple water."; >
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE SimpleWaterVS();
    PixelShader = compile PS_SHADER_TYPE SimpleWaterPS();
    DEFAULT_STATES
  }
}

technique Ocean < string name = "Material still water."; >
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE OceanVS();
    PixelShader = compile PS_SHADER_TYPE StillWaterPS();
    DEFAULT_STATES
  }
}

