#include "constants.fxh"
#include "lighting.fxh"

extern const float4 vRTSize;
extern const float  fFrameTime;

texture2D HDRBackBuffer;
texture2D LastFrameLuminance;
texture2D AdoptedLuminance;

sampler2D samplerHDR = sampler_state
{
  Texture = <HDRBackBuffer>;
  MinFilter = Point;
  MagFilter = Point;
  MipFilter = None;
  AddressU = Clamp;
  AddressV = Clamp;
};

sampler2D samplerLastLum = sampler_state
{
  Texture = <LastFrameLuminance>;
  MinFilter = Point;
  MagFilter = Point;
  MipFilter = None;
  AddressU = Clamp;
  AddressV = Clamp;
};

sampler2D samplerAdoptedLum = sampler_state
{
  Texture = <AdoptedLuminance>;
  MinFilter = Point;
  MagFilter = Point;
  MipFilter = None;
  AddressU = Wrap;
  AddressV = Wrap;
};

struct VS_INPUT
{
	float4 Pos				: POSITION;
	float2 Tex0				: TEXCOORD0;
};

struct vertexOutput
{
  float4 HPosition			: POSITION;
  float2 Tex0		    		: TEXCOORD0;
};

vertexOutput std_VS(const in VS_INPUT In)
{
  vertexOutput Out;
  Out.HPosition = In.Pos;
	Out.HPosition.xy += float2(-.5f, .5f) * vRTSize.zw;
  Out.Tex0 = In.Tex0;
	Out.Tex0.y = 1.f - Out.Tex0.y;

  return Out;
}

half4 InitialLuminancePS(const in vertexOutput IN) : COLOR0
{
	half fLogLumSum = 0.0f;
	static const half2 vOffsets[] = 
	{
		half2(0, 0),
		half2(0, 1),
		half2(1, 0),
		half2(1, 1)
	};

	for(int iSample = 0; iSample < 4; iSample++)
	{
		// Compute the sum of log(luminance) throughout the sample points
		const half3 vSample = tex2D(samplerHDR, IN.Tex0 + vOffsets[iSample] * vRTSize.zw * .5h);
		fLogLumSum += log(Luminance(vSample)+0.0001h);
	}

	// Divide the sum to complete the average
	fLogLumSum /= 4.h;

  return half4(fLogLumSum, fLogLumSum, fLogLumSum, 1.h);
}

float4 EyeAdaptionPS(const in vertexOutput IN) : COLOR0
{
	float fCurrentLum = exp(tex2D(samplerLastLum, float2(0.5f, 0.5f)));

	// check for fp exceptions
	if(isnan(fCurrentLum) || isinf(fCurrentLum))
		fCurrentLum = 1.f;

	// clamp the luminance to the limited range (for limited HDR formats like A2RGB10)
	fCurrentLum = clamp(fCurrentLum, .01f, 32.f);

	// The user's adapted luminance level is simulated by closing the gap between
	// adapted luminance and current luminance by 3% every frame, based on a
	// 30 fps rate. This is not an accurate model of human adaptation, which can
	// take longer than half an hour.
	const float k = saturate(1.f - pow( 0.97f, 30 * max(.001f, fFrameTime) ));
	return float4(fCurrentLum.xxx, k);
}

half4 ToneMappingPS(const in vertexOutput IN) : COLOR0
{
	half4 vSample = tex2D(samplerHDR, IN.Tex0);
	const half fAdaptedLum = tex2D(samplerAdoptedLum, half2(0.5h, 0.5h)).r;

	// Map the high range of color values into a range appropriate for
	// display, taking into account the user's adaptation level, and selected
	// values for for middle gray and white cutoff.
	//static const half fMiddleGray = .25f;       // The middle gray key value
	//vSample.rgb *= fMiddleGray / (fAdaptedLum + 0.001h);
	//vSample.rgb /= (1.0h + vSample);

	// custom exp tone mapper
	half fAdaptedLumDest = 3.h / ( max( 0.1h, 1.h + 5.h * fAdaptedLum ) );
	vSample.xyz = 1 - exp( -fAdaptedLumDest * vSample.xyz ); 

	return vSample;
}

technique InitialLuminance
{
	pass p0
  {
		ZEnable = false;
		CullMode = None;
		VertexShader = compile VS_SHADER_TYPE std_VS();
		PixelShader  = compile PS_SHADER_TYPE InitialLuminancePS();
  }
}

technique EyeAdaption
{
	pass p0
  {
		ZEnable = false;
		CullMode = None;
		AlphaBlendEnable = true;
		SrcBlend = SrcAlpha;
		DestBlend = InvSrcAlpha;
		VertexShader = compile VS_SHADER_TYPE std_VS();
		PixelShader  = compile PS_SHADER_TYPE EyeAdaptionPS();
  }
}

technique ToneMapping
{
	pass p0
  {
		ZEnable = false;
		CullMode = None;
		VertexShader = compile VS_SHADER_TYPE std_VS();
		PixelShader  = compile PS_SHADER_TYPE ToneMappingPS();
  }
}