#include "constants.fxh"
#include "skinning.fxh"
#include "lighting.fxh"

struct vsIn
{
  float4  Position    	: POSITION;
  half3  Normal      	: NORMAL;
#ifdef RT_SKINNING
  float4  BoneID				: BLENDINDICES;
  float4  Weight				: BLENDWEIGHT;
#endif  
#ifdef VERTEX_COLOR
  FIXED4 VtxColor    	  : COLOR0;
#endif
  half2  Tex0        	: TEXCOORD0;
#ifdef TANGENT_BASIS
  half3  Tangent     	: TANGENT;
  half3  Binormal    	: BINORMAL;
#endif
};

struct vsOut
{
  float4 HPosition			: POSITION;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  FIXED4 VtxColor			: COLOR0;
#endif

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW
  FIXED4 FogColor    		: COLOR1;
 #else	
  FIXED FogColor    		  : FOG;
 #endif	
#endif

  half4 Tex0				: TEXCOORD0;
#if RT_PROFILE <= QUALITY_LOW
  FIXED4 Diffuse			: COLOR0;
#endif

#if RT_PROFILE > QUALITY_VERYLOW
  half3 WorldNormal		: TEXCOORD2;
  float4 WorldView		: TEXCOORD3;
  float4 WorldPos	    : TEXCOORD4;
#endif  
#if defined(TANGENT_BASIS) && RT_PROFILE > QUALITY_VERYLOW
  half3 WorldTangent			: TEXCOORD5;
  half3 WorldBinormal		 	: TEXCOORD6;
#endif  
};


half3 GetLDiffuse(const float3 vWorldPos, const float3 vWorldNormal)
{
  half3 vDiffuseAcc = half3(0,0,0);

  float4 WorldLightPos = LGetPosition(0);
  half3 vVertexSpaceLightPos = WorldLightPos.xyz - vWorldPos;
  half3 vDirToLight = normalize(vVertexSpaceLightPos);
  float fNdotL = dot(vDirToLight.xyz, vWorldNormal);
  vDiffuseAcc.rgb = saturate(fNdotL) * LGetDiffuse(0);

  return vDiffuseAcc;
}

void LLighting(inout vsOut Out, const in vsIn In, const float3 vWorldPos, const float3 vWorldNormal)
{
#if RT_PROFILE <= QUALITY_LOW    
  Out.Diffuse = float4(0,0,0,1);
  #ifdef RT_NUM_LIGHTS
    Out.Diffuse.rgb += GetLDiffuse(vWorldPos, vWorldNormal.xyz);
  #endif
  #if defined(VERTEX_COLOR)
   #if !defined(RT_STATIC_LIGHTING)
    Out.Diffuse.rgb *= In.VtxColor.rgb;
   #endif	
    Out.Diffuse.a = In.VtxColor.a;
  #endif
  
  #ifndef RT_LIGHTMAP
    Out.Diffuse.rgb += F_WorldAmbient.rgb;
  #endif
#endif
}

float3 WorldNormal(const in vsIn In, const in float3 Normal) 
{
#ifdef RT_SKINNING
  // cast the vectors to arrays for use in the for loop below
  int4 IndexVector = D3DCOLORtoUBYTE4(In.BoneID);
  float4 BlendWeightsArray = D3DCOLORtoUBYTE4(In.Weight);
  return _SkinningNormal(IndexVector, BlendWeightsArray, Normal);
#else
  return MUL_VEC_MAT(In.Normal.xyz, (float3x3)PI_mW).xyz;
#endif
}

vsOut CharVS(in vsIn In)
{
  vsOut Out = (vsOut)0;

#ifdef RT_SKINNING
  int4 IndexVector = D3DCOLORtoUBYTE4(In.BoneID);
  // cast the vectors to arrays for use in the for loop below
  float4 BlendWeightsArray = D3DCOLORtoUBYTE4(In.Weight); ///255.0f;
  float3 vWorldPos = _SkinningPosition(IndexVector, BlendWeightsArray, In.Position);
#else
  float4 Po = float4(In.Position.xyz, 1);
  float3 vWorldPos = MUL_VEC_MAT(Po, PI_mW).xyz;
#endif
  float3 vWorldNormal = WorldNormal(In, In.Normal);

#if defined(TANGENT_BASIS)
  float3 vWorldTangent = WorldNormal(In, In.Tangent.xyz);
  float3 vWorldBinormal = WorldNormal(In, In.Binormal.xyz);
#endif
   
  Out.Tex0.xy = In.Tex0.xy;

#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  Out.VtxColor = In.VtxColor;
#endif

#if defined(RT_LIGHTMAP) || defined(SECONDARY_DIFFUSE)
  #if defined(RT_CUSTOM_LIGHTMAP) 	 
    Out.Tex0.zw = (vWorldPos.xz - LightmapCoords.xy) * LightmapCoords.z;
    Out.Tex0.zw += LightmapCoords.w;
    Out.Tex0.w = 1 - Out.Tex0.w;
  #else
    half fFracX = (LightmapTRCoords[1].x - vWorldPos.x) / (LightmapTRCoords[1].x - LightmapTRCoords[0].x);
    half fFracY = (vWorldPos.z - LightmapTRCoords[0].y) / (LightmapTRCoords[3].y - LightmapTRCoords[0].y);
    Out.Tex0.zw = lerp(lerp(LightmapTRCoords[0].zw, LightmapTRCoords[1].zw, fFracX), lerp(LightmapTRCoords[2].zw, LightmapTRCoords[3].zw, fFracX), fFracY);
  #endif  
#endif

//#if RT_PROFILE > QUALITY_VERYLOW
  Out.WorldNormal.xyz = vWorldNormal;  
 #if defined(TANGENT_BASIS)
  Out.WorldTangent = vWorldTangent;
  Out.WorldBinormal = vWorldBinormal;
 #endif
  Out.WorldView.xyz = MAT_POS(C_VI) - vWorldPos;
  Out.WorldPos.xyz = vWorldPos;
//#endif

  float4 vWPos = float4(vWorldPos, 1);
  Out.HPosition = MUL_VEC_MAT(vWPos, C_VP);

#if defined (RT_FOG)
 #if RT_PROFILE > QUALITY_VERYLOW && !defined(OPENGL)
  Out.FogColor = Fog(length(Out.WorldView.xyz));
 #else
  Out.FogColor = Fog(Out.HPosition.w);
 #endif	
#endif

#if RT_PROFILE <= QUALITY_LOW
  //LLighting(Out, In, vWorldPos, vWorldNormal);
#endif

  return Out;
}

FIXED4 CharPS(in vsOut IN) : COLOR
{
  half3 vNormal = IN.WorldNormal.xyz;
  half2 baseTC = IN.Tex0.xy;
  half2 lmTC = IN.Tex0.zw;
  half2 bumpTC = baseTC;

  half3 mTangentToWS0, mTangentToWS1, mTangentToWS2;
#if defined (TANGENT_BASIS)
  mTangentToWS0 = IN.WorldTangent.xyz;  
  mTangentToWS1 = IN.WorldBinormal.xyz;  
  mTangentToWS2 = IN.WorldNormal.xyz;  
#endif

  FIXED4 cDiffuseMap = tex2D(ColorSampler, baseTC);

  FIXED fAlpha = cDiffuseMap.a;
  FIXED4 cVertexColor = 1;
#if defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  cVertexColor = IN.VtxColor;
  fAlpha *= cVertexColor.a;
#endif

#if defined(SECONDARY_DIFFUSE) && defined(VERTEX_COLOR) && RT_PROFILE > QUALITY_LOW
  half4 vSecondColor = tex2D(SecondColorSampler, lmTC);
  float fA = cVertexColor.a;
  cDiffuseMap = lerp(cDiffuseMap, vSecondColor, fA);
#endif

#if defined(BUMP_DIFFUSE) && (defined(RT_SPECULAR) || defined (RT_ENVMAP))
    half4 cBumpData = tex2D(SpecularSampler, bumpTC);
    half3 cBumpMap = cBumpData.rgb * 2.h - 1.h;
  #if RT_PROFILE > QUALITY_HIGH
    cBumpMap = normalize(cBumpMap);
  #endif
  #if defined(TANGENT_BASIS)
    vNormal.xyz  = cBumpMap.x * mTangentToWS0;
    vNormal.xyz += cBumpMap.y * mTangentToWS1;
    vNormal.xyz += cBumpMap.z * mTangentToWS2;
  #endif
#endif
 

  float3 vView = normalize(IN.WorldView.xyz);
  float fNdotE = dot(vView, vNormal);
  float3 vReflVec = 2 * fNdotE * vNormal - vView;

  FIXED3 cDiffuseAcc = 0;
  FIXED3 cAmbientAcc = 0;
  FIXED3 cSpecularAcc = 0;
  FIXED fSpecularMap = 1;

  FIXED3 cLM = 1;
#if defined(RT_LIGHTMAP)
  cLM = tex2D(LightMapSampler, lmTC.xy).rgb * 2.h;
#endif

//#if RT_PROFILE > QUALITY_LOW
  cAmbientAcc = F_WorldAmbient.rgb;
//#endif
#if defined(RT_LIGHTMAP)
  cAmbientAcc = cLM;
#endif

//#if RT_PROFILE <= QUALITY_LOW
//  cDiffuseAcc = IN.Diffuse.rgb;
//#else

#if defined(RT_NUM_LIGHTS)

 #if defined(BUMP_DIFFUSE) && (defined(RT_SPECULAR) || defined (RT_ENVMAP))
  fSpecularMap *= cBumpData.a;
 #endif

  // Assume this is sun
  FIXED3 cDiffuse = LGetDiffuse(0);
  FIXED3 cSpecular = LGetSpecular(0);

  half3 vDirToLight;
  FIXED fFallOff;
  GetLighting(vDirToLight, fFallOff, IN.WorldPos.xyz, LIGHT_DIRECTIONAL, 0, 0);
  half fNdotL = dot(vDirToLight, vNormal);
  cDiffuseAcc += cDiffuse * fNdotL;

 #if defined(RT_SPECULAR)
  half fSpecPow = 8; 
#if RT_PROFILE >= QUALITY_MEDIUM
  fSpecPow = max(4.h, Glossiness);
#endif
  half fSpec = Phong(vNormal, vView, vDirToLight, fSpecPow);
  cSpecular *= fSpec * fSpecularMap * Ks;
  cSpecularAcc += cSpecular * cLM;
 #endif
#endif

  
#if RT_PROFILE > QUALITY_VERYLOW
  // affect specular by a Hemisphere term
  //cSpecularAcc += saturate(vNormal.y * 2.h) * Ks;
#endif

#if defined (RT_ENVMAP) && RT_PROFILE > QUALITY_VERYLOW
  half fFresnelTerm = 1;
  FIXED3 vEnv = texCUBE(EnvMapSampler, vReflVec).rgb;
 #if defined (RT_FRESNEL)
  fFresnelTerm = GetFresnel(fNdotE, RT_PROFILE, true);
 #endif
  FIXED3 cEnv = vEnv * fFresnelTerm * Ks * fSpecularMap;
  cSpecularAcc += cEnv;
  fAlpha *= fFresnelTerm;
#endif

#if defined(RT_LIGHTMAP) && defined(RT_CUSTOM_LIGHTMAP) && defined(RT_SPECULAR) && RT_PROFILE > QUALITY_VERYLOW
  cSpecularAcc.xyz *= cAmbientAcc.xyz;
#endif

  FIXED4 cOut;
  cOut.xyz = (cAmbientAcc.xyz + cDiffuseAcc.xyz) * cDiffuseMap.xyz * cVertexColor.rgb + cSpecularAcc;
  cOut.w = fAlpha;

#if defined(RT_FOG) && RT_PROFILE > QUALITY_VERYLOW
  frag_fog_setup(IN.FogColor, cOut);
#endif

#if defined (RT_ENVMAP) && RT_PROFILE > QUALITY_VERYLOW
  //cOut.xyz = cEnv.xyz;
#endif

  return cOut;
}

technique Default
{
  pass p0
  {
    VertexShader = compile VS_SHADER_TYPE CharVS();
    PixelShader = compile PS_SHADER_TYPE CharPS();
    DEFAULT_STATES
  }
}
