//General lighting constants

#ifndef _LIGHTING_DECLARED__
#define _LIGHTING_DECLARED__

#include "shadowing.fxh"

half Luminance(const in half3 color)
{
	half3 vLumVector = half3(0.2125h, 0.7154h, 0.0721h);
	return dot(color, vLumVector);
}


//  Setup fog. Per-vertex for alpha blended objects and per-object volumetric fog
#if (RT_PROFILE > QUALITY_VERYLOW)
half4	Fog(const in half sceneZ)
#else
half	Fog(const in half sceneZ)
#endif
{
  half effDist = max(0, (sceneZ - F_FogParams.x)*F_FogParams.y);
	half density = exp(-effDist*effDist);
	half3 fogColor = lerp(F_FogColorEnd.xyz, F_FogColor.xyz, density);
	half fogDensity = saturate(F_FogColorEnd.w*density + F_FogColor.w);
#ifdef RT_BLEND_ADD
  fogColor = half3(0,0,0);
#endif

#if (RT_PROFILE > QUALITY_VERYLOW)
  return half4(fogColor, fogDensity);
#else
 #if !defined(RT_BLEND_ALPHA) && !defined(RT_BLEND_ADD)
  return saturate(exp(-effDist*effDist));
 #else
  return 1;
 #endif 
#endif  
}

#if defined(RT_FOG) && (RT_PROFILE > QUALITY_VERYLOW)
void frag_fog_setup(const in half4 FogColor, inout FIXED4 cOut)
{
  half3 FogC = FogColor.xyz;
 #if defined(RT_BLEND_MULTI)
   FogC = 1;
 #elif defined(RT_BLEND_ALPHA)
   cOut.w *= FogColor.w;
   FogC = 0;
 #endif
	cOut.xyz = lerp(FogC.xyz, cOut.xyz, FogColor.w);
}
#else
	#define frag_fog_setup(FogColor, cOut)
#endif

#define LIGHT_UNKNOWN     0
#define LIGHT_POINT       1
#define LIGHT_SPOT        2
#define LIGHT_DIRECTIONAL 3

half3 LGetDiffuse(int nID)
{
  if (nID == 0)
    return L_LightDiffuse0.xyz;
  if (nID == 1)
    return L_LightDiffuse1.xyz;
  if (nID == 2)
    return L_LightDiffuse2.xyz;
  if (nID == 3)
    return L_LightDiffuse3.xyz;
  return half3(0,0,0);
}

half3 LGetSpecular(int nID)
{
  if (nID == 0)
    return L_LightDiffuse0.xyz;
  if (nID == 1)
    return L_LightDiffuse1.xyz;
  if (nID == 2)
    return L_LightDiffuse2.xyz;
  if (nID == 3)
    return L_LightDiffuse3.xyz;
  return half3(0,0,0);
}

float4 LGetPosition(int nID)
{
  if (nID == 0)
    return L_LightPos0;
  if (nID == 1)
    return L_LightPos1;
  if (nID == 2)
    return L_LightPos2;
  if (nID == 3)
    return L_LightPos3;
  return float4(0,0,0,1);
}

half4 LGetDir(int nID)
{
  if (nID == 0)
    return L_LightDir0;
  if (nID == 1)
    return L_LightDir1;
  if (nID == 2)
    return L_LightDir2;
  if (nID == 3)
    return L_LightDir3;
  return float4(0,0,0,1);
}

FIXED4 LGetShadowMask(int nID)
{
  if (nID == 0)
    return FIXED4(1,0,0,0);
  if (nID == 1)
    return FIXED4(0,1,0,0);
  if (nID == 2)
    return FIXED4(0,0,1,0);
  if (nID == 3)
    return FIXED4(0,0,0,1);
  return FIXED4(0,0,0,0);
}

void GetLighting(out half3 vDirToLight, out FIXED fFallOff, const in float3 vVertexPos, int nLightType, int iLightIndex, int nTerrain)
{
	if(nLightType == LIGHT_DIRECTIONAL)
	{
		vDirToLight = LGetDir(iLightIndex).rgb;
#if defined(RT_LIGHTMAP) && !defined(RT_PLANAR_LIGHTMAP) && !defined(VERTEX_COLOR)
		//fFallOff = 0.h;	// disable sun for lightmaps
		fFallOff = 1.h;	// disable sun for lightmaps
#else
		fFallOff = 1.h;	// disable attenuation for sun
#endif
    if (nTerrain)
      fFallOff = 1.h;	// disable attenuation for sun
		return;
	}

	float4 WorldLightPos = LGetPosition(iLightIndex);
	float3 vVertexSpaceLightPos = WorldLightPos.xyz - vVertexPos;
	vDirToLight = normalize(vVertexSpaceLightPos);

	// Quadratic distance attenuation
	fFallOff = 1.h - saturate(dot(vVertexSpaceLightPos, vVertexSpaceLightPos) * WorldLightPos.w);

	if (nLightType == LIGHT_SPOT)
	{
		half4 vDirCosPhi = LGetDir(iLightIndex);
		fFallOff *= saturate((-dot(vDirToLight, vDirCosPhi.xyz) - vDirCosPhi.w) * 10.f);
	}
} 

half GetFresnel(in half NdotE, int nShaderQuality, bool bDoubleSided)
{
  if (bDoubleSided)
    NdotE = abs(NdotE);

#ifndef OPENGL
  if (nShaderQuality < QUALITY_LOW)
    return 1.h;
#endif
  half fPow = FresnelBiasScalePow.z;
  if (nShaderQuality <= QUALITY_MEDIUM)
    fPow = 4.0h;

  half fFacing = saturate(1.h - NdotE);
  half fFresnel = FresnelBiasScalePow.x + FresnelBiasScalePow.y * pow(fFacing, fPow);
  
  return fFresnel;
}

//Light equations

//Blinn-Phong Lighting 
//point light

void BlinnPhong(const in half3 LightColor,
		    const in half3 Nn,
		    const in half3 Ln,
		    const in half3 Hn,
        const in half SpecExp,
		    out half3 DiffuseContrib,
		    out half3 SpecularContrib)
{
  half2 litV = saturate(half2( dot(Ln, Nn), pow(saturate(dot(Hn, Nn)), SpecExp)));
  DiffuseContrib = litV.x * LightColor;
  SpecularContrib = litV.x * litV.y * LightColor;
}

void BlinnPhongAtt(const in half3 LightColor, const in half3 LightAtt, const in half3 Dist,
		    const in half3 Nn,
		    const in half3 Ln,
		    const in half3 Hn,
        const in half SpecExp,
		    out half3 DiffuseContrib,
		    out half3 SpecularContrib)
{
  half  att = 1.h / dot(LightAtt, Dist);
  half2 litV = saturate(float2( dot(Ln, Nn), pow(saturate(dot(Hn, Nn)), SpecExp)))*att;
  DiffuseContrib = litV.x * LightColor;
  SpecularContrib = litV.x * litV.y * LightColor;
}

half LeafShadingBack(const in half3 vEye, const in half3 vLight, const in half3 vNormal)
{            
  half EdotL = saturate(dot(vEye, -vLight));          
  // Tweaked NdotL wrapping
  half fLdotNBack = saturate(dot(-vNormal, vLight) * .3h + .7h);
  half powEdotL = EdotL * EdotL;
  powEdotL *= powEdotL;
  return lerp(powEdotL, fLdotNBack, .8h);
}

void BlinnPhongDS(const in half3 LightColor,
		    const in half3 Nn,
		    const in half3 Ln,
		    const in half3 Hn,
        const in half SpecExp,
		    out half3 DiffuseContrib,
		    out half3 SpecularContrib)
{
  half LdN = dot(Ln, Nn);
  half spec = saturate(pow(saturate(dot(Hn, Nn)), SpecExp));
  DiffuseContrib = abs(LdN) * LightColor;
  SpecularContrib = LdN * spec * LightColor;
}

// pseudo Lambert fake-translusent organic
// LambertLight (rgb - subsurface color a - diffuse intensity)

void AnisoLight(const in half3 LightColor,
		    const in half3 Nn,
		    const in half3 Ln,
		    const in half3 Hn,
        const in sampler2D texAniso,
		    out half3 DiffuseContrib,
		    out half3 SpecularContrib)
{
	half NdotL = max(0, dot(Ln, Nn));
  half4 tex = tex2D(texAniso, half2(NdotL, dot(Hn, Nn)));
  DiffuseContrib = NdotL * LightColor;
  SpecularContrib = (tex.rgb * tex.aaa) * 4.h * LightColor;
}

//atmosphere scattering
void Scattering(const in half3 vViewN, const in half sceneZ, out half4 Inscatter,out half4 Extinction) 
{
	half4 scattering;
#if defined(RT_NUM_LIGHTS)
	scattering.x = dot(vViewN, LGetDir(0).xyz);
#else
	scattering.x = 1.h;
#endif
	scattering.y = 1.h + scattering.x*scattering.x;
	scattering.z = sceneZ; 

  float log2_e = 1.f / log(2.f);

	half3 extinction = F_BetaRayleighMie * scattering.z * log2_e;
	extinction = exp(-extinction);

	// phase2
	scattering.w = F_HG.x / (pow(F_HG.y + F_HG.z * scattering.x, (3.h/2.h)));

	// Inscattering (I) = (Beta'_1 * Phase_1(theta) + Beta'_2 * Phase_2(theta)) * [1-exp(-Beta_1*s).exp(-Beta_2*s)] / (Beta_1 + Beta_2)
	half3 inscattering = (F_BetaDashRayleigh.xyz*scattering.y);
	inscattering += (F_BetaDashMie.xyz*scattering.w);
	inscattering *= (1.h-extinction);
	inscattering *= F_OneOverBetaRM.xyz;

	// Apply Inscattering contribution factors.
	inscattering *= F_TermMultiplier.y;

	// Scale with Sun color & intesity.
	inscattering *= F_SunColorIntensity.xyz;
	inscattering *= F_SunColorIntensity.w;

	extinction *= F_SunColorIntensity.xyz;
	extinction *= F_SunColorIntensity.w*.01h;

	Inscatter.xyz = inscattering;
	Inscatter.a = 1.h;

	Extinction.xyz = extinction;
	Extinction.a = 1.h;            
}

//////////////////////////////// Phong model /////////////////////////////////////
// - Phong model has good properties for plastic and some metallic surfaces. 
// - Good for general use. Very cheap.

half Phong(half3 N, half3 V, half3 L, half Exp)
{
  half3 R = reflect(-L, N);					// 3 ALU
  return pow( max(0.01h, saturate( dot(V, R) )) ,  max(0.01h, Exp));	// 4
  
  // 7 ALU
}

// Optimized phong, use if mirrowed reflection vector pre-computed
half Phong(half3 R, half3 L, half Exp)
{	
  return pow( saturate( dot(L, R) ) ,  Exp); //4
  // 4 ALU
}

//////////////////////////////// Ward model, anisotropic /////////////////////////
// - Ward anisotropic model has good properties for generic anisotropic highlights 
// (can have radial or disk shape)


half WardAnisotropic_opt(half3 T, half3 B, half3 N, half3 V, half3 L, half NdotL, half NdotE, half2 vSurfaceRoughness, half Exp)
{
  // opt: averaging vectors instead of normalization
  half3 H =  normalize(V + L);                                        // 1 alu
  half NdotH = dot( N, H );                                                    // 1 alu
  half2 HdotT_B = ( half2( dot( H, B ), dot( H, T ) ) )* vSurfaceRoughness.xy;                 // 3 alu
  HdotT_B *= HdotT_B;                                                                             // 1 alu
  // using sat instead, since it's free, but compensating by brighting specular
  half specK = saturate( rsqrt( saturate( NdotL * NdotE ) ) );                                 // 2 alu
  // opt: merged exponential with power factor
  specK *= NdotL * exp( dot(HdotT_B.xy, -2.h * Exp/(1.001h + NdotH)) );  // 6 alu
  return specK;
}

#endif // _LIGHTING_DECLARED__
