// Terrain
#include "lighting.fxh"
#include "fragLib.fxh"


#define MAX_SHADOW_SAMPLES_NUM 8
extern const float4 vPoissonCoeffs[MAX_SHADOW_SAMPLES_NUM/2];
extern const float4x4 mShadowWLP0;

#define SMAP_SIZE 512

texture2D ShadowMap;
texture2D JitterTexture;

sampler2D SamplerShadow = sampler_state {
	Texture = <ShadowMap>;
	minfilter = Point;
	magfilter = Point;
	mipfilter = None;
	AddressU = Clamp;
	AddressV = Clamp;
};
    
sampler2D SamplerJitter = sampler_state {
	Texture = <JitterTexture>;
	minfilter = Point;
	magfilter = Point;
	mipfilter = None;
	AddressU = Wrap;
	AddressV = Wrap;
};

sampler2D SamplerTexture = sampler_state {
	Texture = <ShadowMap>;
	minfilter = Linear;
	magfilter = Linear;
	mipfilter = None;
	AddressU = Clamp;
	AddressV = Clamp;
};

struct VSOUTPUT_LAND
{
	float4 HPosition 				: POSITION;
	float4 ShadowCoord      : TEXCOORD0;	
};

#define SHADOW_EPSILON 0.00005f

struct VS_INPUT
{
	float4 vPos					: POSITION;
};

struct VS_INPUT_TC
{
	float4 Pos					: POSITION;
	float2 tc0          : TEXCOORD0;	
};

struct VS_OUTPUT_TC
{
	float4 HPosition 				: POSITION;
	float2 tc0              : TEXCOORD0;	
};

VSOUTPUT_LAND VS(const in VS_INPUT In)
{
	VSOUTPUT_LAND Out;

  Out.HPosition = mul(float4(In.vPos.xzy, 1), C_VP);
  Out.ShadowCoord = mul(float4(In.vPos.xyz, 1), mShadowWLP0);

	return Out;
}


float4 PS(const in VSOUTPUT_LAND In):COLOR
{
  float fZ = In.ShadowCoord.z / In.ShadowCoord.w - 0.001; //SHADOW_EPSILON;
  float fOffset = In.ShadowCoord.w / SMAP_SIZE;
  float fShadowAmount = 0;
    
  float fBias = 0.007999;
  float fDepthShift = 0.5;
  
  // Variance shadow fetching
  float4 p = In.ShadowCoord;
  p.xy /= p.w;
  p.z *= C_FarClip.y;
  p.z -= fBias;
  float4 moments = tex2D(SamplerShadow, p.xy);
  
	// Variance shadow mapping
  float M = moments.r; //mean

	float E_x2 = moments.g;
	float Ex_2 = M * M;
	float variance = (E_x2 - Ex_2)/16; //decrease range of variance to increase precision for 16-bit formats

  M += fDepthShift; 

  float m_d = M - p.z;
	float p_max = variance / (variance + m_d * m_d);

	// Standard shadow map comparison
	float lit_factor = m_d > 0;

	//select properly shadow region because of one-tailed version of inequality
	fShadowAmount = max(lit_factor, p_max);

  float fShadow[9];  
  fShadow[0] = 0.0625f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(-fOffset, -fOffset, 0, 0) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[1] = 0.125f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(0,        -fOffset, 0, 0) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[2] = 0.0625f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(fOffset,  -fOffset, 0, 0) ) < fZ) ? 0.0f : 1.0f);  

  fShadow[3] = 0.125f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(-fOffset, 0, 0, 0) ) < fZ) ? 0.0f : 1.0f);
  fShadow[4] = 0.25f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(0,        0, 0, 0) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[5] = 0.125f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(fOffset,  0, 0, 0) ) < fZ) ? 0.0f : 1.0f);  

  fShadow[6] = 0.0625f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(-fOffset, fOffset, 0, 0) ) < fZ) ? 0.0f : 1.0f);
  fShadow[7] = 0.125f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(0,        fOffset, 0, 0) ) < fZ) ? 0.0f : 1.0f);  
  fShadow[8] = 0.0625f * ((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(fOffset,  fOffset, 0, 0) ) < fZ) ? 0.0f : 1.0f);  

  fShadowAmount = fShadow[0]+fShadow[1]+fShadow[2]+fShadow[3]+fShadow[4]+fShadow[5]+fShadow[6]+fShadow[7]+fShadow[8];
  //fShadowAmount = 1; //((tex2Dproj( SamplerShadow, In.ShadowCoord + float4(0,        0, 0, 0) ) < fZ) ? 0.0f : 1.0f);  
  
  return float4(fShadowAmount, fShadowAmount, fShadowAmount, fShadowAmount);
}

VS_OUTPUT_TC default_VS(const in VS_INPUT_TC In)
{
  VS_OUTPUT_TC Out;
  Out.HPosition = In.Pos;
  Out.tc0 = In.tc0;
  return Out;
}

float4 downsample_PS(const in VS_OUTPUT_TC In) : COLOR0
{
  float2 Offset = F_ScreenSize.xy;
  
  float4 Shadow[9];
  
  Shadow[0] = 0.0625f * tex2D( SamplerTexture, In.tc0 + float2(-Offset.x, -Offset.y) );
  Shadow[1] = 0.125f * tex2D( SamplerTexture, In.tc0 + float2(0,        -Offset.y) );
  Shadow[2] = 0.0625f * tex2D( SamplerTexture, In.tc0 + float2(Offset.x,  -Offset.y) );

  Shadow[3] = 0.125f * tex2D( SamplerTexture, In.tc0 + float2(-Offset.x, 0) );
  Shadow[4] = 0.25f * tex2D( SamplerTexture, In.tc0 + float2(0,        0) );
  Shadow[5] = 0.125f * tex2D( SamplerTexture, In.tc0 + float2(Offset.x,  0) );

  Shadow[6] = 0.0625f * tex2D( SamplerTexture, In.tc0 + float2(-Offset.x, Offset.y) );
  Shadow[7] = 0.125f * tex2D( SamplerTexture, In.tc0 + float2(0,        Offset.y) );
  Shadow[8] = 0.0625f * tex2D( SamplerTexture, In.tc0 + float2(Offset.x,  Offset.y) );

  float4 sum = Shadow[0]+Shadow[1]+Shadow[2]+Shadow[3]+Shadow[4]+Shadow[5]+Shadow[6]+Shadow[7]+Shadow[8];
  
  return sum;
}

VS_OUTPUT_TC buildmask_VS(const in VS_INPUT_TC In)
{
  VS_OUTPUT_TC Out;
  Out.HPosition = In.Pos;
  Out.tc0 = In.tc0;
  return Out;
}

float4 buildmask_PS(const in VS_OUTPUT_TC In) : COLOR0
{
  float4 vColor = tex2D(SamplerTexture, In.tc0);

  return vColor;
}

technique TerrainLM
{
	pass P0
	{
		VertexShader = compile VS_SHADER_TYPE VS();
		PixelShader = compile PS_SHADER_TYPE PS();
		AlphaBlendEnable = false;
	}
}

technique TerrainLMBlur
{
	pass P0
	{
		VertexShader = compile VS_SHADER_TYPE default_VS();
		PixelShader = compile PS_SHADER_TYPE downsample_PS();
		AlphaBlendEnable = false;
	}
}

technique TerrainBuildLMMask
{
	pass P0
	{
		VertexShader = compile VS_SHADER_TYPE buildmask_VS();
		PixelShader = compile PS_SHADER_TYPE buildmask_PS();
		AlphaBlendEnable = false;
	}
}
